"""
elasticapm.contrib.flask
~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import absolute_import

import logging
import os
import warnings

import flask
from flask import request, signals

import elasticapm.instrumentation.control
from elasticapm.base import Client
from elasticapm.conf import setup_logging
from elasticapm.contrib.flask.utils import get_data_from_request
from elasticapm.handlers.logging import LoggingHandler
from elasticapm.utils import (build_name_with_http_method_prefix,
                              disabled_due_to_debug)
from elasticapm.utils.deprecation import deprecated

logger = logging.getLogger('elasticapm.errors.client')


def make_client(client_cls, app, app_name=None, secret_token=None):
    config = app.config.get('ELASTICAPM', {})
    app_name = (
        app_name or
        config.get('APP_NAME') or  # config
        os.environ.get('ELASTICAPM_APP_NAME') # environment
    )
    secret_token = (
        secret_token or
        config.get('SECRET_TOKEN') or  # config
        os.environ.get('ELASTICAPM_SECRET_TOKEN') # environment
    )
    if hasattr(flask, '__version__'):
        framework_version = 'flask/' + flask.__version__
    else:
        framework_version = 'flask/<0.7'

    client = client_cls(
        app_name=app_name,
        secret_token=secret_token,
        include_paths=set(config.get('INCLUDE_PATHS', [])) | set([app.import_name]),
        exclude_paths=config.get('EXCLUDE_PATHS'),
        filter_exception_types=config.get('FILTER_EXCEPTION_TYPES', None),
        servers=config.get('SERVERS'),
        transport_class=config.get('TRANSPORT_CLASS', None),
        hostname=config.get('HOSTNAME'),
        auto_log_stacks=config.get('AUTO_LOG_STACKS'),
        timeout=config.get('TIMEOUT'),
        string_max_length=config.get('STRING_MAX_LENGTH'),
        list_max_length=config.get('LIST_MAX_LENGTH'),
        traces_freq_send=config.get('TRACES_FREQ_SEND'),
        processors=config.get('PROCESSORS'),
        async_mode=config.get('ASYNC_MODE'),
        transactions_ignore_patterns=config.get('TRANSACTIONS_IGNORE_PATTERNS'),
    )

    client._framework = 'flask'
    client._framework_version = getattr(flask, '__version__', '<0.7')
    return client


class ElasticAPM(object):
    """
    Flask application for Elastic APM.

    Look up configuration from ``os.environ.get('ELASTICAPM_APP_NAME')`` and
    ``os.environ.get('ELASTICAPM_SECRET_TOKEN')``::

    >>> elasticapm = ElasticAPM(app)

    Pass an arbitrary APP_NAME and SECRET_TOKEN::

    >>> elasticapm = ElasticAPM(app, app_name='myapp', secret_token='asdasdasd')

    Pass an explicit client::

    >>> elasticapm = ElasticAPM(app, client=client)

    Automatically configure logging::

    >>> elasticapm = ElasticAPM(app, logging=True)

    Capture an exception::

    >>> try:
    >>>     1 / 0
    >>> except ZeroDivisionError:
    >>>     elasticapm.capture_exception()

    Capture a message::

    >>> elasticapm.captureMessage('hello, world!')
    """
    def __init__(self, app=None, app_name=None,
                 secret_token=None, client=None, client_cls=Client,
                 logging=False):
        self.app_name = app_name
        self.secret_token = secret_token
        self.logging = logging
        self.client_cls = client_cls
        self.client = client

        if app:
            self.init_app(app)

    def handle_exception(self, *args, **kwargs):
        if not self.client:
            return

        if disabled_due_to_debug(
            self.app.config.get('ELASTICAPM', {}),
            self.app.config.get('DEBUG', False)
        ):
            return

        self.client.capture(
            'Exception', exc_info=kwargs.get('exc_info'),
            data={'context': {'request': get_data_from_request(request)}},
            extra={
                'app': self.app,
            },
        )

    def init_app(self, app):
        self.app = app
        if not self.client:
            self.client = make_client(
                self.client_cls,
                app,
                self.app_name,
                self.secret_token,
            )

        if self.logging:
            setup_logging(LoggingHandler(self.client))

        signals.got_request_exception.connect(self.handle_exception, sender=app, weak=False)

        # Instrument to get traces
        skip_env_var = 'SKIP_INSTRUMENT'
        if skip_env_var in os.environ:
            logger.debug("Skipping instrumentation. %s is set.", skip_env_var)
        else:
            elasticapm.instrumentation.control.instrument()

            signals.request_started.connect(self.request_started, sender=app)
            signals.request_finished.connect(self.request_finished, sender=app)

    def request_started(self, app):
        self.client.begin_transaction("web.flask")

    def request_finished(self, app, response):
        rule = request.url_rule.rule if request.url_rule is not None else ""
        rule = build_name_with_http_method_prefix(rule, request)

        self.client.end_transaction(rule, response.status_code)

    def capture_exception(self, *args, **kwargs):
        assert self.client, 'capture_exception called before application configured'
        return self.client.capture_exception(*args, **kwargs)

    def capture_message(self, *args, **kwargs):
        assert self.client, 'capture_message called before application configured'
        return self.client.capture_message(*args, **kwargs)

    @deprecated(alternative="capture_exception()")
    def captureException(self, *args, **kwargs):
        return self.capture_exception(*args, **kwargs)

    @deprecated(alternative="capture_message()")
    def captureMessage(self, *args, **kwargs):
        return self.capture_message(*args, **kwargs)



import datetime
import functools
import logging
import re
import threading
import time
import uuid

from elasticapm.conf import defaults
from elasticapm.utils import get_name_from_func
from elasticapm.utils.lru import LRUCache

__all__ = ('trace', )

error_logger = logging.getLogger('elasticapm.errors')

thread_local = threading.local()
thread_local.transaction = None


_time_func = time.time


def get_transaction(clear=False):
    """
    Get the transaction registered for the current thread.

    :return:
    :rtype: Transaction
    """
    transaction = getattr(thread_local, "transaction", None)
    if clear:
        thread_local.transaction = None
    return transaction


class Transaction(object):
    _lrucache = LRUCache(maxsize=5000)

    def __init__(self, get_frames, kind="custom"):
        self.id = uuid.uuid4()
        self.timestamp = datetime.datetime.utcnow()
        self.start_time = _time_func()
        self.name = None
        self.duration = None
        self.result = None
        self.kind = kind
        self.get_frames = get_frames

        self.traces = []
        self.trace_stack = []
        self.ignore_subtree = False
        self.extra = {}

        self._trace_counter = 0

        # The transaction is a trace as well
        self.begin_trace("transaction", "transaction")

    def end_transaction(self, skip_frames=8):
        # End the "transaction" trace started above
        self.duration = _time_func() - self.start_time
        return self.end_trace(skip_frames)

    def begin_trace(self, name, trace_type, context=None, leaf=False):
        # If we were already called with `leaf=True`, we'll just push
        # a placeholder on the stack.
        if self.ignore_subtree:
            self.trace_stack.append(None)
            return None

        if leaf:
            self.ignore_subtree = True

        start = _time_func() - self.start_time
        trace = Trace(self._trace_counter, name, trace_type, start, context)
        self._trace_counter += 1
        self.trace_stack.append(trace)
        return trace

    def end_trace(self, skip_frames):
        trace = self.trace_stack.pop()
        if trace is None:
            return None

        self.ignore_subtree = False

        trace.duration = _time_func() - trace.start_time - self.start_time

        if self.trace_stack:
            trace.parent = self.trace_stack[-1].idx

        trace.frames = self.get_frames()[skip_frames:]
        self.traces.append(trace)

        return trace

    def to_dict(self):
        return {
            'id': str(self.id),
            'name': self.name,
            'type': self.kind,
            'duration': self.duration * 1000,  # milliseconds
            'result': str(self.result),
            'timestamp': self.timestamp.strftime(defaults.TIMESTAMP_FORMAT),
            'context': self.extra,
            'traces': [
                trace_obj.to_dict() for trace_obj in self.traces
            ]
        }


class Trace(object):
    def __init__(self, idx, name, trace_type, start_time, context=None,
                 leaf=False):
        """
        Create a new trace

        :param idx: Index of this trace
        :param name: Generic name of the trace
        :param trace_type: type of the trace
        :param start_time: start time relative to the transaction
        :param context: context dictionary
        :param leaf: is this transaction a leaf transaction?
        """
        self.idx = idx
        self.name = name
        self.type = trace_type
        self.context = context
        self.leaf = leaf
        self.start_time = start_time
        self.duration = None
        self.transaction = None
        self.parent = None
        self.frames = None

    @property
    def fingerprint(self):
        return self.transaction, self.parent, self.name, self.type

    def to_dict(self):
        return {
            'id': self.idx,
            'name': self.name,
            'type': self.type,
            'start': self.start_time * 1000,  # milliseconds
            'duration': self.duration * 1000,  # milliseconds
            'parent': self.parent,
            'stacktrace': self.frames,
            'context': self.context
        }


class TransactionsStore(object):
    def __init__(self, get_frames, collect_frequency, ignore_patterns=None):
        self.cond = threading.Condition()
        self.collect_frequency = collect_frequency
        self._get_frames = get_frames
        self._transactions = []
        self._last_collect = _time_func()
        self._ignore_patterns = [re.compile(p) for p in ignore_patterns or []]

    def add_transaction(self, transaction):
        with self.cond:
            self._transactions.append(transaction)
            self.cond.notify()

    def get_all(self, blocking=False):
        with self.cond:
            # If blocking is true, always return at least 1 item
            while blocking and len(self._transactions) == 0:
                self.cond.wait()
            transactions, self._transactions = self._transactions, []
        self._last_collect = _time_func()
        return transactions

    def should_collect(self):
        return (_time_func() - self._last_collect) >= self.collect_frequency

    def __len__(self):
        with self.cond:
            return len(self._transactions)

    def begin_transaction(self, transaction_type):
        """
        Start a new transactions and bind it in a thread-local variable

        :returns the Transaction object
        """
        transaction = Transaction(self._get_frames, transaction_type)
        thread_local.transaction = transaction
        return transaction

    def _should_ignore(self, transaction_name):
        for pattern in self._ignore_patterns:
            if pattern.search(transaction_name):
                return True
        return False

    def end_transaction(self, response_code, transaction_name):
        transaction = get_transaction(clear=True)
        if transaction:
            transaction.end_transaction()
            if self._should_ignore(transaction_name):
                return
            transaction.name = transaction_name
            transaction.result = response_code
            self.add_transaction(transaction.to_dict())
        return transaction


class trace(object):
    def __init__(self, name=None, trace_type='code.custom', extra=None,
                 skip_frames=0, leaf=False):
        self.name = name
        self.type = trace_type
        self.extra = extra
        self.skip_frames = skip_frames
        self.leaf = leaf

    def __call__(self, func):
        self.name = self.name or get_name_from_func(func)

        @functools.wraps(func)
        def decorated(*args, **kwds):
            with self:
                return func(*args, **kwds)

        return decorated

    def __enter__(self):
        transaction = get_transaction()

        if transaction:
            transaction.begin_trace(self.name, self.type, self.extra,
                                    self.leaf)

    def __exit__(self, exc_type, exc_val, exc_tb):
        transaction = get_transaction()

        if transaction:
            transaction.end_trace(self.skip_frames)



import time

from mock import Mock

from elasticapm.traces import TransactionsStore, get_transaction, trace
from tests.utils.compat import TestCase


class RequestStoreTest(TestCase):
    def setUp(self):
        self.mock_get_frames = Mock()

        frames = [{'function': 'something_expensive',
                   'abs_path': '/var/parent-elasticapm/elasticapm/tests/contrib/django/testapp/views.py',
                   'lineno': 52, 'module': 'tests.contrib.django.testapp.views',
                   'filename': 'tests/contrib/django/testapp/views.py'},
                  {'function': '_resolve_lookup',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/template/base.py',
                   'lineno': 789, 'module': 'django.template.base',
                   'filename': 'django/template/base.py'},
                  {'function': 'resolve',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/template/base.py',
                   'lineno': 735, 'module': 'django.template.base',
                   'filename': 'django/template/base.py'},
                  {'function': 'resolve',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/template/base.py',
                   'lineno': 585, 'module': 'django.template.base',
                   'filename': 'django/template/base.py'}, {'lineno': 4,
                                                            'filename': u'/var/parent-elasticapm/elasticapm/tests/contrib/django/testapp/templates/list_fish.html'},
                  {'function': 'render',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/template/defaulttags.py',
                   'lineno': 4, 'module': 'django.template.defaulttags',
                   'filename': 'django/template/defaulttags.py'},
                  {'function': 'render_node',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/template/debug.py',
                   'lineno': 78, 'module': 'django.template.debug',
                   'filename': 'django/template/debug.py'},
                  {'function': 'render',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/template/base.py',
                   'lineno': 840, 'module': 'django.template.base',
                   'filename': 'django/template/base.py'},
                  {'function': 'instrumented_test_render',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/test/utils.py',
                   'lineno': 85, 'module': 'django.test.utils',
                   'filename': 'django/test/utils.py'}, {'function': 'render',
                                                         'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/template/base.py',
                                                         'lineno': 140,
                                                         'module': 'django.template.base',
                                                         'filename': 'django/template/base.py'},
                  {'function': 'rendered_content',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/template/response.py',
                   'lineno': 82, 'module': 'django.template.response',
                   'filename': 'django/template/response.py'},
                  {'function': 'render',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/template/response.py',
                   'lineno': 105, 'module': 'django.template.response',
                   'filename': 'django/template/response.py'},
                  {'function': 'get_response',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/core/handlers/base.py',
                   'lineno': 137, 'module': 'django.core.handlers.base',
                   'filename': 'django/core/handlers/base.py'},
                  {'function': '__call__',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/test/client.py',
                   'lineno': 109, 'module': 'django.test.client',
                   'filename': 'django/test/client.py'}, {'function': 'request',
                                                          'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/test/client.py',
                                                          'lineno': 426,
                                                          'module': 'django.test.client',
                                                          'filename': 'django/test/client.py'},
                  {'function': 'get',
                   'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/test/client.py',
                   'lineno': 280, 'module': 'django.test.client',
                   'filename': 'django/test/client.py'}, {'function': 'get',
                                                          'abs_path': '/home/ron/.virtualenvs/elasticapm/local/lib/python2.7/site-packages/django/test/client.py',
                                                          'lineno': 473,
                                                          'module': 'django.test.client',
                                                          'filename': 'django/test/client.py'},
                  {'function': 'test_template_name_as_view',
                   'abs_path': '/var/parent-elasticapm/elasticapm/tests/contrib/django/django_tests.py',
                   'lineno': 710, 'module': 'tests.contrib.django.django_tests',
                   'filename': 'tests/contrib/django/django_tests.py'}]

        self.mock_get_frames.return_value = frames
        self.requests_store = TransactionsStore(self.mock_get_frames, 99999)

    def test_lru_get_frames_cache(self):
        self.requests_store.begin_transaction("transaction.test")

        for i in range(10):
            with trace("bleh", "custom"):
                time.sleep(0.01)

        self.assertEqual(self.mock_get_frames.call_count, 10)

    def test_leaf_tracing(self):
        self.requests_store.begin_transaction("transaction.test")

        with trace("root", "custom"):
            with trace("child1-leaf", "custom", leaf=True):

                # These two traces should not show up
                with trace("ignored-child1", "custom", leaf=True):
                    time.sleep(0.01)

                with trace("ignored-child2", "custom", leaf=False):
                    time.sleep(0.01)

        self.requests_store.end_transaction(None, "transaction")

        transactions = self.requests_store.get_all()
        traces = transactions[0]['traces']

        self.assertEqual(len(traces), 3)

        signatures = ['transaction', 'root', 'child1-leaf']
        self.assertEqual(set([t['name'] for t in traces]),
                         set(signatures))


def test_get_transaction():
    requests_store = TransactionsStore(lambda: [], 99999)
    t = requests_store.begin_transaction("test")
    assert t == get_transaction()


def test_get_transaction_clear():
    requests_store = TransactionsStore(lambda: [], 99999)
    t = requests_store.begin_transaction("test")
    assert t == get_transaction(clear=True)
    assert get_transaction() is None



# -*- coding: utf-8 -*-

from elasticapm.transport.exceptions import InvalidScheme


class TransportException(Exception):
    def __init__(self, message, data=None, print_trace=True):
        super(TransportException, self).__init__(message)
        self.data = data
        self.print_trace = print_trace


class Transport(object):
    """
    All transport implementations need to subclass this class

    You must implement a send method..
    """
    async_mode = False
    scheme = []

    def check_scheme(self, url):
        if url.scheme not in self.scheme:
            raise InvalidScheme()

    def send(self, data, headers):
        """
        You need to override this to do something with the actual
        data. Usually - this is sending to a server
        """
        raise NotImplementedError

    def close(self):
        """
        Cleans up resources and closes connection
        :return:
        """
        pass


class AsyncTransport(Transport):
    async_mode = True

    def send_async(self, data, headers, success_callback=None, fail_callback=None):
        raise NotImplementedError



import asyncio
import urllib.parse

from elasticapm.base import Client


class Client(Client):

    def handle_transport_response(self, task):
        try:
            url = task.result()
        except Exception as exc:
            self.handle_transport_fail(exception=exc)
        else:
            self.handle_transport_success(url=url)

    def _send_remote(self, url, data, headers=None):
        if headers is None:
            headers = {}
        parsed = urllib.parse.urlparse(url)
        transport = self._get_transport(parsed)
        loop = asyncio.get_event_loop()
        task = loop.create_task(
            transport.send(data, headers, timeout=self.timeout))
        task.add_done_callback(self.handle_transport_response)



from elasticapm.instrumentation.packages.dbapi2 import (ConnectionProxy,
                                                        CursorProxy,
                                                        DbApi2Instrumentation,
                                                        extract_signature)


class MySQLCursorProxy(CursorProxy):
    provider_name = 'mysql'

    def extract_signature(self, sql):
        return extract_signature(sql)


class MySQLConnectionProxy(ConnectionProxy):
    cursor_proxy = MySQLCursorProxy


class MySQLInstrumentation(DbApi2Instrumentation):
    name = 'mysql'

    instrument_list = [
        ("MySQLdb", "connect"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        return MySQLConnectionProxy(wrapped(*args, **kwargs))



# -*- coding: utf-8 -*-
import logging
import socket

from elasticapm.conf import defaults
from elasticapm.contrib.async_worker import AsyncWorker
from elasticapm.transport.base import (AsyncTransport, Transport,
                                       TransportException)
from elasticapm.utils.compat import HTTPError

try:
    from urllib2 import Request, urlopen
except ImportError:
    from urllib.request import Request, urlopen


logger = logging.getLogger('elasticapm')


class HTTPTransport(Transport):

    scheme = ['http', 'https']

    def __init__(self, parsed_url):
        self.check_scheme(parsed_url)

        self._parsed_url = parsed_url
        self._url = parsed_url.geturl()

    def send(self, data, headers, timeout=None):
        """
        Sends a request to a remote webserver using HTTP POST.

        Returns the shortcut URL of the recorded error on Elastic APM
        """
        req = Request(self._url, headers=headers)
        if timeout is None:
            timeout = defaults.TIMEOUT
        response = None
        try:
            try:
                response = urlopen(req, data, timeout)
            except TypeError:
                response = urlopen(req, data)
        except Exception as e:
            print_trace = True
            if isinstance(e, socket.timeout):
                message = (
                    "Connection to APM Server timed out "
                    "(url: %s, timeout: %d seconds)" % (self._url, timeout)
                )
            elif isinstance(e, HTTPError):
                body = e.read()
                if e.code == 429:  # rate-limited
                    message = 'Temporarily rate limited: '
                    print_trace = False
                else:
                    message = 'Unable to reach APM Server: '
                message += '%s (url: %s, body: %s)' % (e, self._url, body)
            else:
                message = 'Unable to reach APM Server: %s (url: %s)' % (
                    e, self._url
                )
            raise TransportException(message, data, print_trace=print_trace)
        finally:
            if response:
                response.close()

        return response.info().get('Location')


class AsyncHTTPTransport(AsyncTransport, HTTPTransport):
    scheme = ['http', 'https']
    async_mode = True

    def __init__(self, parsed_url):
        super(AsyncHTTPTransport, self).__init__(parsed_url)
        if self._url.startswith('async+'):
            self._url = self._url[6:]
        self._worker = None

    @property
    def worker(self):
        if not self._worker or not self._worker.is_alive():
            self._worker = AsyncWorker()
        return self._worker

    def send_sync(self, data=None, headers=None, success_callback=None,
                  fail_callback=None):
        try:
            url = HTTPTransport.send(self, data, headers)
            if callable(success_callback):
                success_callback(url=url)
        except Exception as e:
            if callable(fail_callback):
                fail_callback(exception=e)

    def send_async(self, data, headers, success_callback=None,
                   fail_callback=None):
        kwargs = {
            'data': data,
            'headers': headers,
            'success_callback': success_callback,
            'fail_callback': fail_callback,
        }
        self.worker.queue(self.send_sync, kwargs)

    def close(self):
        if self._worker:
            self._worker.main_thread_terminated()



# -*- coding: utf-8 -*-
import os

import certifi
import urllib3
from urllib3.exceptions import MaxRetryError, TimeoutError

from elasticapm.conf import defaults
from elasticapm.transport.base import TransportException
from elasticapm.transport.http import AsyncHTTPTransport, HTTPTransport


class Urllib3Transport(HTTPTransport):

    scheme = ['http', 'https']

    def __init__(self, parsed_url):
        kwargs = {
            'cert_reqs': 'CERT_REQUIRED',
            'ca_certs': certifi.where(),
            'block': True,
        }
        proxy_url = os.environ.get('HTTPS_PROXY', os.environ.get('HTTP_PROXY'))
        if proxy_url:
            self.http = urllib3.ProxyManager(proxy_url, **kwargs)
        else:
            self.http = urllib3.PoolManager(**kwargs)
        super(Urllib3Transport, self).__init__(parsed_url)

    def send(self, data, headers, timeout=None):
        if timeout is None:
            timeout = defaults.TIMEOUT
        response = None
        try:
            try:
                response = self.http.urlopen(
                    'POST', self._url, body=data, headers=headers, timeout=timeout
                )
            except Exception as e:
                print_trace = True
                if isinstance(e, MaxRetryError) and isinstance(e.reason, TimeoutError):
                    message = (
                        "Connection to APM Server timed out "
                        "(url: %s, timeout: %d seconds)" % (self._url, timeout)
                    )
                    print_trace = False
                else:
                    message = 'Unable to reach APM Server: %s (url: %s)' % (
                        e, self._url
                    )
                raise TransportException(message, data, print_trace=print_trace)
            body = response.read()
            if response.status >= 400:
                if response.status == 429:  # rate-limited
                    message = 'Temporarily rate limited: '
                    print_trace = False
                else:
                    message = 'HTTP %s: ' % response.status
                    print_trace = True
                message += body.decode('utf8')
                raise TransportException(message, data, print_trace=print_trace)
            return response.getheader('Location')
        finally:
            if response:
                response.close()


class AsyncUrllib3Transport(AsyncHTTPTransport, Urllib3Transport):
    scheme = ['http', 'https']
    async_mode = True

    def send_sync(self, data=None, headers=None, success_callback=None,
                  fail_callback=None):
        try:
            url = Urllib3Transport.send(self, data, headers)
            if callable(success_callback):
                success_callback(url=url)
        except Exception as e:
            if callable(fail_callback):
                fail_callback(exception=e)



import pytest  # isort:skip
pytest.importorskip("django")  # isort:skip

from os.path import join

import django
from django.core.urlresolvers import reverse
from django.test import TestCase

import mock
import pytest

from conftest import BASE_TEMPLATE_DIR
from elasticapm.contrib.django.models import elasticapm, get_client

# Testing Django 1.8+ backends
TEMPLATES = (
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            BASE_TEMPLATE_DIR
        ],
    },
    {
        'BACKEND': 'django.template.backends.jinja2.Jinja2',
        'DIRS': [
            join(BASE_TEMPLATE_DIR, 'jinja2')
        ],
    },
)


class TracesTest(TestCase):
    def setUp(self):
        self.elasticapm_client = get_client()
        elasticapm.instrumentation.control.instrument()

    @mock.patch("elasticapm.traces.TransactionsStore.should_collect")
    def test_template_rendering(self, should_collect):
        should_collect.return_value = False
        with self.settings(MIDDLEWARE_CLASSES=[
            'elasticapm.contrib.django.middleware.TracingMiddleware']):
            self.client.get(reverse('render-heavy-template'))
            self.client.get(reverse('render-heavy-template'))
            self.client.get(reverse('render-heavy-template'))

        transactions = self.elasticapm_client.instrumentation_store.get_all()

        self.assertEqual(len(transactions), 3)
        traces = transactions[0]['traces']
        self.assertEqual(len(traces), 3, [t['name'] for t in traces])

        kinds = ['transaction', 'code', 'template.django']
        self.assertEqual(set([t['type'] for t in traces]),
                         set(kinds))

        self.assertEqual(traces[0]['type'], 'code')
        self.assertEqual(traces[0]['name'], 'something_expensive')
        self.assertEqual(traces[0]['parent'], 1)

        self.assertEqual(traces[1]['type'], 'template.django')
        self.assertEqual(traces[1]['name'], 'list_users.html')
        self.assertEqual(traces[1]['parent'], 0)

        self.assertEqual(traces[2]['type'], 'transaction')
        self.assertEqual(traces[2]['name'], 'transaction')
        self.assertIsNone(traces[2]['parent'])

    @pytest.mark.skipif(django.VERSION < (1, 8),
                        reason='Jinja2 support introduced with Django 1.8')
    @mock.patch("elasticapm.traces.TransactionsStore.should_collect")
    def test_template_rendering_django18_jinja2(self, should_collect):
        should_collect.return_value = False
        with self.settings(MIDDLEWARE_CLASSES=[
                'elasticapm.contrib.django.middleware.TracingMiddleware'],
                TEMPLATES=TEMPLATES
            ):
            self.client.get(reverse('render-jinja2-template'))
            self.client.get(reverse('render-jinja2-template'))
            self.client.get(reverse('render-jinja2-template'))

        transactions = self.elasticapm_client.instrumentation_store.get_all()

        self.assertEqual(len(transactions), 3)
        traces = transactions[0]['traces']
        self.assertEqual(len(traces), 2, [t['name'] for t in traces])

        kinds = ['transaction', 'template.jinja2']
        self.assertEqual(set([t['type'] for t in traces]),
                         set(kinds))

        self.assertEqual(traces[0]['type'], 'template.jinja2')
        self.assertEqual(traces[0]['name'], 'jinja2_template.html')
        self.assertEqual(traces[0]['parent'], 0)

        self.assertEqual(traces[1]['type'], 'transaction')
        self.assertEqual(traces[1]['name'], 'transaction')
        self.assertIsNone(traces[1]['parent'])



from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import trace


class Jinja2Instrumentation(AbstractInstrumentedModule):
    name = 'jinja2'

    instrument_list = [
        ("jinja2", "Template.render"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        signature = instance.name or instance.filename
        with trace(signature, "template.jinja2"):
            return wrapped(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from mock import Mock

from elasticapm.events import Message
from tests.utils.compat import TestCase


class MessageTest(TestCase):
    def test_to_string(self):
        unformatted_message = 'My message from %s about %s'
        formatted_message = unformatted_message % (1, 2)
        client = Mock()
        message = Message()
        message.logger = Mock()
        data = {
            'log': {
                'message': unformatted_message % (1, 2),
                'param_message': unformatted_message,
            }
        }

        self.assertEqual(message.to_string(client, data), formatted_message)



from functools import partial

import mock
import pytest
import redis
from redis.client import StrictRedis

import elasticapm
import elasticapm.instrumentation.control
from elasticapm.traces import trace
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


@pytest.mark.integrationtest
class InstrumentRedisTest(TestCase):
    def setUp(self):
        self.client = get_tempstoreclient()
        elasticapm.instrumentation.control.instrument()

    @mock.patch("elasticapm.traces.TransactionsStore.should_collect")
    def test_pipeline(self, should_collect):
        should_collect.return_value = False
        self.client.begin_transaction("transaction.test")
        with trace("test_pipeline", "test"):
            conn = redis.StrictRedis()
            pipeline = conn.pipeline()
            pipeline.rpush("mykey", "a", "b")
            pipeline.expire("mykey", 1000)
            pipeline.execute()
        self.client.end_transaction("MyView")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']

        expected_signatures = ['transaction', 'test_pipeline',
                               'StrictPipeline.execute']

        self.assertEqual(set([t['name'] for t in traces]),
                         set(expected_signatures))

        self.assertEqual(traces[0]['name'], 'StrictPipeline.execute')
        self.assertEqual(traces[0]['type'], 'cache.redis')

        self.assertEqual(traces[1]['name'], 'test_pipeline')
        self.assertEqual(traces[1]['type'], 'test')

        self.assertEqual(traces[2]['name'], 'transaction')
        self.assertEqual(traces[2]['type'], 'transaction')

        self.assertEqual(len(traces), 3)

    @mock.patch("elasticapm.traces.TransactionsStore.should_collect")
    def test_rq_patches_redis(self, should_collect):
        should_collect.return_value = False

        # Let's go ahead and change how something important works
        conn = redis.StrictRedis()
        conn._pipeline = partial(StrictRedis.pipeline, conn)

        self.client.begin_transaction("transaction.test")
        with trace("test_pipeline", "test"):
            # conn = redis.StrictRedis()
            pipeline = conn._pipeline()
            pipeline.rpush("mykey", "a", "b")
            pipeline.expire("mykey", 1000)
            pipeline.execute()
        self.client.end_transaction("MyView")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']

        expected_signatures = ['transaction', 'test_pipeline',
                               'StrictPipeline.execute']

        self.assertEqual(set([t['name'] for t in traces]),
                         set(expected_signatures))

        self.assertEqual(traces[0]['name'], 'StrictPipeline.execute')
        self.assertEqual(traces[0]['type'], 'cache.redis')

        self.assertEqual(traces[1]['name'], 'test_pipeline')
        self.assertEqual(traces[1]['type'], 'test')

        self.assertEqual(traces[2]['name'], 'transaction')
        self.assertEqual(traces[2]['type'], 'transaction')

        self.assertEqual(len(traces), 3)

    @mock.patch("elasticapm.traces.TransactionsStore.should_collect")
    def test_redis_client(self, should_collect):
        should_collect.return_value = False
        self.client.begin_transaction("transaction.test")
        with trace("test_redis_client", "test"):
            conn = redis.StrictRedis()
            conn.rpush("mykey", "a", "b")
            conn.expire("mykey", 1000)
        self.client.end_transaction("MyView")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']

        expected_signatures = ['transaction', 'test_redis_client',
                               'RPUSH', 'EXPIRE']

        self.assertEqual(set([t['name'] for t in traces]),
                         set(expected_signatures))

        self.assertEqual(traces[0]['name'], 'RPUSH')
        self.assertEqual(traces[0]['type'], 'cache.redis')

        self.assertEqual(traces[1]['name'], 'EXPIRE')
        self.assertEqual(traces[1]['type'], 'cache.redis')

        self.assertEqual(traces[2]['name'], 'test_redis_client')
        self.assertEqual(traces[2]['type'], 'test')

        self.assertEqual(traces[3]['name'], 'transaction')
        self.assertEqual(traces[3]['type'], 'transaction')

        self.assertEqual(len(traces), 4)



"""
elasticapm.events
~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import logging
import sys
import uuid

from elasticapm.utils import varmap
from elasticapm.utils.encoding import shorten, to_unicode
from elasticapm.utils.stacks import (get_culprit, get_stack_info,
                                     iter_traceback_frames)

__all__ = ('BaseEvent', 'Exception', 'Message', 'Query')

logger = logging.getLogger(__name__)


class BaseEvent(object):
    @staticmethod
    def to_string(client, data):
        raise NotImplementedError

    @staticmethod
    def capture(client, **kwargs):
        return {}


class Exception(BaseEvent):
    """
    Exceptions store the following metadata:

    - value: 'My exception value'
    - type: 'ClassName'
    - module '__builtin__' (i.e. __builtin__.TypeError)
    - frames: a list of serialized frames (see _get_traceback_frames)
    """

    @staticmethod
    def to_string(client, data):
        exc = data['exception']
        if exc['value']:
            return '%s: %s' % (exc['type'], exc['value'])
        return exc['type']

    @staticmethod
    def get_hash(data):
        exc = data['exception']
        output = [exc['type']]
        for frame in data['stacktrace']['frames']:
            output.append(frame['module'])
            output.append(frame['function'])
        return output

    @staticmethod
    def capture(client, exc_info=None, **kwargs):
        culprit = exc_value = exc_type = exc_module = frames = None
        new_exc_info = False
        if not exc_info or exc_info is True:
            new_exc_info = True
            exc_info = sys.exc_info()

        if not exc_info:
            raise ValueError('No exception found')

        try:
            exc_type, exc_value, exc_traceback = exc_info

            frames = varmap(
                lambda k, v: shorten(v,
                                     string_length=client.string_max_length,
                                     list_length=client.list_max_length
                                     ),
                get_stack_info((iter_traceback_frames(exc_traceback)))
            )

            culprit = get_culprit(frames, client.include_paths,
                                  client.exclude_paths)

            if hasattr(exc_type, '__module__'):
                exc_module = exc_type.__module__
                exc_type = exc_type.__name__
            else:
                exc_module = None
                exc_type = exc_type.__name__
        finally:
            if new_exc_info:
                try:
                    del exc_info
                    del exc_traceback
                except Exception as e:
                    logger.exception(e)

        return {
            'id': str(uuid.uuid4()),
            'culprit': culprit,
            'exception': {
                'message': '%s: %s' % (exc_type, to_unicode(exc_value)) if exc_value else str(exc_type),
                'type': str(exc_type),
                'module': str(exc_module),
                'stacktrace':  frames,
            }
        }


class Message(BaseEvent):
    """
    Messages store the following metadata:

    - message: 'My message from %s about %s'
    - params: ('foo', 'bar')
    """

    @staticmethod
    def to_string(client, data):
        return data['log']['message']

    @staticmethod
    def get_hash(data):
        msg = data['param_message']
        return [msg['message']]

    @staticmethod
    def capture(client, param_message=None, message=None, **kwargs):
        if message:
            param_message = {'message': message}
        params = param_message.get('params', ())
        message = param_message['message'] % params
        data = kwargs.get('data', {})
        message_data = {
            'id': str(uuid.uuid4()),
            'log': {
                'level': data.get('level', 'error'),
                'logger_name': data.get('logger'),
                'message': message,
                'param_message': param_message['message'],
            }
        }
        if isinstance(data.get('stacktrace'), dict):
            message_data['log']['stacktrace'] = data['stacktrace']['frames']
        return message_data


class Query(BaseEvent):
    """
    Messages store the following metadata:

    - query: 'SELECT * FROM table'
    - engine: 'postgesql_psycopg2'
    """

    def to_string(self, data):
        sql = data['query']
        return sql['query']

    def get_hash(self, data):
        sql = data['query']
        return [sql['query'], sql['engine']]

    def capture(self, query, engine, **kwargs):
        return {
            'query': {
                'query': query,
                'engine': engine,
            }
        }



import boto3
import mock

import elasticapm
import elasticapm.instrumentation.control
from elasticapm.traces import trace
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


class InstrumentBotocoreTest(TestCase):
    def setUp(self):
        self.client = get_tempstoreclient()
        elasticapm.instrumentation.control.instrument()

    @mock.patch("botocore.endpoint.Endpoint.make_request")
    def test_botocore_instrumentation(self, mock_make_request):
        mock_response = mock.Mock()
        mock_response.status_code = 200
        mock_make_request.return_value = (mock_response, {})

        self.client.begin_transaction("transaction.test")
        with trace("test_pipeline", "test"):
            session = boto3.Session(aws_access_key_id='foo',
                                    aws_secret_access_key='bar',
                                    region_name='us-west-2')
            ec2 = session.client('ec2')
            ec2.describe_instances()
        self.client.end_transaction("MyView")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']
        self.assertIn('ec2:DescribeInstances', map(lambda x: x['name'], traces))



"""
elasticapm.handlers.logging
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import absolute_import

import datetime
import logging
import sys
import traceback

from elasticapm.base import Client
from elasticapm.utils import six
from elasticapm.utils.encoding import to_string
from elasticapm.utils.stacks import iter_stack_frames


class LoggingHandler(logging.Handler, object):
    def __init__(self, *args, **kwargs):
        client = kwargs.pop('client_cls', Client)
        if len(args) == 1:
            arg = args[0]
            args = args[1:]
            if isinstance(arg, Client):
                self.client = arg
            else:
                raise ValueError(
                    'The first argument to %s must be a Client instance, '
                    'got %r instead.' % (
                        self.__class__.__name__,
                        arg,
                    ))
        elif 'client' in kwargs:
            self.client = kwargs.pop('client')
        else:
            self.client = client(*args, **kwargs)

        super(LoggingHandler, self).__init__(*args, **kwargs)

    def emit(self, record):
        self.format(record)

        # Avoid typical config issues by overriding loggers behavior
        if record.name.startswith('elasticapm.errors'):
            six.print_(to_string(record.message), file=sys.stderr)
            return

        try:
            return self._emit(record)
        except Exception:
            six.print_(
                "Top level ElasticAPM exception caught - "
                "failed creating log record",
                sys.stderr)
            six.print_(to_string(record.msg), sys.stderr)
            six.print_(to_string(traceback.format_exc()), sys.stderr)

            try:
                self.client.capture('Exception')
            except Exception:
                pass

    def _emit(self, record, **kwargs):
        data = {}

        for k, v in six.iteritems(record.__dict__):
            if '.' not in k and k not in ('culprit',):
                continue
            data[k] = v

        stack = getattr(record, 'stack', None)
        if stack is True:
            stack = iter_stack_frames()

        if stack:
            frames = []
            started = False
            last_mod = ''
            for item in stack:
                if isinstance(item, (list, tuple)):
                    frame, lineno = item
                else:
                    frame, lineno = item, item.f_lineno

                if not started:
                    f_globals = getattr(frame, 'f_globals', {})
                    module_name = f_globals.get('__name__', '')
                    if last_mod.startswith(
                            'logging') and not module_name.startswith(
                            'logging'):
                        started = True
                    else:
                        last_mod = module_name
                        continue
                frames.append((frame, lineno))
            stack = frames

        extra = getattr(record, 'data', {})
        # Add in all of the data from the record that we aren't already capturing
        for k in record.__dict__.keys():
            if k in (
                    'stack', 'name', 'args', 'msg', 'levelno', 'exc_text',
                    'exc_info', 'data', 'created', 'levelname', 'msecs',
                    'relativeCreated'):
                continue
            if k.startswith('_'):
                continue
            extra[k] = record.__dict__[k]

        date = datetime.datetime.utcfromtimestamp(record.created)

        # If there's no exception being processed,
        # exc_info may be a 3-tuple of None
        # http://docs.python.org/library/sys.html#sys.exc_info
        if record.exc_info and all(record.exc_info):
            handler = self.client.get_handler('elasticapm.events.Exception')

            data.update(handler.capture(self.client, exc_info=record.exc_info))
            # data['checksum'] = handler.get_hash(data)

        data['level'] = record.levelno
        data['logger'] = record.name
        return self.client.capture('Message',
                                   param_message={'message': record.msg,
                                                  'params': record.args},
                                   stack=stack, data=data, extra=extra,
                                   date=date, **kwargs)



"""
elasticapm.contrib.django
~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

default_app_config = 'elasticapm.contrib.django.apps.ElasticAPMConfig'

from elasticapm.contrib.django.client import *



from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import trace
from elasticapm.utils import default_ports


class Urllib3Instrumentation(AbstractInstrumentedModule):
    name = 'urllib3'

    instrument_list = [
        ("urllib3.connectionpool", "HTTPConnectionPool.urlopen"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        if 'method' in kwargs:
            method = kwargs['method']
        else:
            method = args[0]

        host = instance.host

        if instance.port != default_ports.get(instance.scheme):
            host += ":" + str(instance.port)

        if 'url' in kwargs:
            url = kwargs['url']
        else:
            url = args[1]

        signature = method.upper() + " " + host

        url = instance.scheme + "://" + host + url

        with trace(signature, "ext.http.urllib3",
                                       {'url': url}, leaf=True):
            return wrapped(*args, **kwargs)



import functools
import warnings

# https://wiki.python.org/moin/PythonDecoratorLibrary#Smart_deprecation_warnings_.28with_valid_filenames.2C_line_numbers.2C_etc..29
# Updated to work with 2.6 and 3+.
from elasticapm.utils import six


def deprecated(alternative=None):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""
    def real_decorator(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            msg = "Call to deprecated function {0}.".format(func.__name__)
            if alternative:
                msg += " Use {0} instead".format(alternative)
            warnings.warn_explicit(
                msg,
                category=DeprecationWarning,
                filename=six.get_function_code(func).co_filename,
                lineno=six.get_function_code(func).co_firstlineno + 1
            )
            return func(*args, **kwargs)
        return new_func
    return real_decorator



"""Provides classes to instrument dbapi2 providers

https://www.python.org/dev/peps/pep-0249/
"""
import re

from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import trace
from elasticapm.utils import six, wrapt


class Literal(object):
    def __init__(self, literal_type, content):
        self.literal_type = literal_type
        self.content = content

    def __eq__(self, other):
        return (isinstance(other, Literal)
                and self.literal_type == other.literal_type
                and self.content == other.content)

    def __repr__(self):
        return "<Literal {}{}{}>".format(self.literal_type, self.content,
                                         self.literal_type)


def skip_to(start, tokens, value_sequence):
    i = start
    while i < len(tokens):
        for idx, token in enumerate(value_sequence):
            if tokens[i+idx] != token:
                break
        else:
            # Match
            return tokens[start:i+len(value_sequence)]
        i += 1

    # Not found
    return None


def look_for_table(sql, keyword):
    tokens = tokenize(sql)
    table_name = _scan_for_table_with_tokens(tokens, keyword)
    if isinstance(table_name, Literal):
        table_name = table_name.content.strip(table_name.literal_type)
    return table_name


def _scan_for_table_with_tokens(tokens, keyword):
    seen_keyword = False
    for idx, lexeme in scan(tokens):
        if seen_keyword:
            if lexeme == "(":
                return _scan_for_table_with_tokens(tokens[idx:], keyword)
            else:
                return lexeme

        if isinstance(lexeme, six.string_types) and lexeme.upper() == keyword:
            seen_keyword = True


def tokenize(sql):
    return [t for t in re.split("(\W)", sql) if t != '']


def scan(tokens):
    literal_start_idx = None
    literal_started = None
    prev_was_escape = False
    lexeme = []

    i = 0
    while i < len(tokens):
        token = tokens[i]
        if literal_start_idx:
            if prev_was_escape:
                prev_was_escape = False
                lexeme.append(token)
            else:

                if token == literal_started:
                    if (literal_started == "'" and len(tokens) > i+1
                            and tokens[i+1] == "'"):  # double quotes
                        i += 1
                        lexeme.append("'")
                    else:
                        yield i, Literal(literal_started, "".join(lexeme))
                        literal_start_idx = None
                        literal_started = None
                        lexeme = []
                else:
                    if token == '\\':
                        prev_was_escape = token
                    else:
                        prev_was_escape = False
                        lexeme.append(token)
        elif literal_start_idx is None:
            if token in ["'", '"', "`"]:
                literal_start_idx = i
                literal_started = token
            elif token == "$":
                # Postgres can use arbitrary characters between two $'s as a
                # literal separation token, e.g.: $fish$ literal $fish$
                # This part will detect that and skip over the literal.
                skipped_token = skip_to(i+1, tokens, '$')
                if skipped_token is not None:
                    dollar_token = ['$'] + skipped_token

                    skipped = skip_to(i + len(dollar_token), tokens,
                                      dollar_token)
                    if skipped:  # end wasn't found.
                        yield i, Literal("".join(dollar_token),
                                         "".join(skipped[:-len(dollar_token)]))
                        i = i + len(skipped) + len(dollar_token)
            else:
                if token != ' ':
                    yield i, token
        i += 1

    if lexeme:
        yield i, lexeme


def extract_signature(sql):
    sql = sql.strip()
    first_space = sql.find(' ')
    if first_space < 0:
        return sql

    second_space = sql.find(' ', first_space+1)

    sql_type = sql[0:first_space].upper()

    if sql_type in ['INSERT', 'DELETE']:
        keyword = 'INTO' if sql_type == 'INSERT' else 'FROM'
        sql_type = sql_type + " " + keyword

        table_name = look_for_table(sql, keyword)
    elif sql_type in ['CREATE', 'DROP']:
        # 2nd word is part of SQL type
        sql_type = sql_type + sql[first_space:second_space]
        table_name = ''
    elif sql_type == 'UPDATE':
        table_name = look_for_table(sql, "UPDATE")
    elif sql_type == 'SELECT':
        # Name is first table
        try:
            sql_type = 'SELECT FROM'
            table_name = look_for_table(sql, "FROM")
        except:
            table_name = ''
    else:
        # No name
        table_name = ''

    signature = ' '.join(filter(bool, [sql_type, table_name]))
    return signature


class CursorProxy(wrapt.ObjectProxy):
    provider_name = None

    def callproc(self, procname, params=None):
        return self._trace_sql(self.__wrapped__.callproc, procname,
                               params)

    def execute(self, sql, params=None):
        return self._trace_sql(self.__wrapped__.execute, sql, params)

    def executemany(self, sql, param_list):
        return self._trace_sql(self.__wrapped__.executemany, sql,
                               param_list)

    def _trace_sql(self, method, sql, params):
        signature = self.extract_signature(sql)
        kind = "db.{0}.sql".format(self.provider_name)
        with trace(signature, kind, {"sql": sql}):
            if params is None:
                return method(sql)
            else:
                return method(sql, params)

    def extract_signature(self, sql):
        raise NotImplementedError()


class ConnectionProxy(wrapt.ObjectProxy):
    cursor_proxy = CursorProxy

    def cursor(self, *args, **kwargs):
        return self.cursor_proxy(self.__wrapped__.cursor(*args, **kwargs))


class DbApi2Instrumentation(AbstractInstrumentedModule):
    connect_method = None

    def call(self, module, method, wrapped, instance, args, kwargs):
        return ConnectionProxy(wrapped(*args, **kwargs))

    def call_if_sampling(self, module, method, wrapped, instance, args, kwargs):
        # Contrasting to the superclass implementation, we *always* want to
        # return a proxied connection, even if there is no ongoing elasticapm
        # transaction yet. This ensures that we instrument the cursor once
        # the transaction started.
        return self.call(module, method, wrapped, instance, args, kwargs)



from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import trace


class PyMongoInstrumentation(AbstractInstrumentedModule):
    name = 'pymongo'

    instrument_list = [
        ("pymongo.collection", "Collection.aggregate"),
        ("pymongo.collection", "Collection.bulk_write"),
        ("pymongo.collection", "Collection.count"),
        ("pymongo.collection", "Collection.create_index"),
        ("pymongo.collection", "Collection.create_indexes"),
        ("pymongo.collection", "Collection.delete_many"),
        ("pymongo.collection", "Collection.delete_one"),
        ("pymongo.collection", "Collection.distinct"),
        ("pymongo.collection", "Collection.drop"),
        ("pymongo.collection", "Collection.drop_index"),
        ("pymongo.collection", "Collection.drop_indexes"),
        ("pymongo.collection", "Collection.ensure_index"),
        ("pymongo.collection", "Collection.find_and_modify"),
        ("pymongo.collection", "Collection.find_one"),
        ("pymongo.collection", "Collection.find_one_and_delete"),
        ("pymongo.collection", "Collection.find_one_and_replace"),
        ("pymongo.collection", "Collection.find_one_and_update"),
        ("pymongo.collection", "Collection.group"),
        ("pymongo.collection", "Collection.inline_map_reduce"),
        ("pymongo.collection", "Collection.insert"),
        ("pymongo.collection", "Collection.insert_many"),
        ("pymongo.collection", "Collection.insert_one"),
        ("pymongo.collection", "Collection.map_reduce"),
        ("pymongo.collection", "Collection.reindex"),
        ("pymongo.collection", "Collection.remove"),
        ("pymongo.collection", "Collection.rename"),
        ("pymongo.collection", "Collection.replace_one"),
        ("pymongo.collection", "Collection.save"),
        ("pymongo.collection", "Collection.update"),
        ("pymongo.collection", "Collection.update_many"),
        ("pymongo.collection", "Collection.update_one"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        cls_name, method_name = method.split('.', 1)
        signature = '.'.join([instance.full_name, method_name])
        with trace(signature, "db.mongodb.query", leaf=True):
            return wrapped(*args, **kwargs)


class PyMongoBulkInstrumentation(AbstractInstrumentedModule):
    name = 'pymongo'

    instrument_list = [
        ("pymongo.bulk", "BulkOperationBuilder.execute"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        collection = instance._BulkOperationBuilder__bulk.collection
        signature = '.'.join([collection.full_name, 'bulk.execute'])
        with trace(signature, "db.mongodb.query"):
            return wrapped(*args, **kwargs)


class PyMongoCursorInstrumentation(AbstractInstrumentedModule):
    name = 'pymongo'

    instrument_list = [
        ("pymongo.cursor", "Cursor._refresh"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        collection = instance.collection
        signature = '.'.join([collection.full_name, 'cursor.refresh'])
        with trace(signature, "db.mongodb.query"):
            return wrapped(*args, **kwargs)



from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import trace


class PythonMemcachedInstrumentation(AbstractInstrumentedModule):
    name = 'python_memcached'

    method_list = [
        'add',
        'append',
        'cas',
        'decr',
        'delete',
        'delete_multi',
        'disconnect_all',
        'flush_all',
        'get',
        'get_multi',
        'get_slabs',
        'get_stats',
        'gets',
        'incr',
        'prepend',
        'replace',
        'set',
        'set_multi',
        'touch'
    ]
    # Took out 'set_servers', 'reset_cas', 'debuglog', 'check_key' and
    # 'forget_dead_hosts' because they involve no communication.

    def get_instrument_list(self):
        return [("memcache", "Client." + method) for method in self.method_list]

    def call(self, module, method, wrapped, instance, args, kwargs):
        name = self.get_wrapped_name(wrapped, instance, method)

        with trace(name, "cache.memcached"):
            return wrapped(*args, **kwargs)



from __future__ import absolute_import

import logging

from django.contrib.auth.models import User
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, render, render_to_response

from elasticapm.traces import trace


class IgnoredException(Exception):
    skip_elasticapm = True


def no_error(request):
    return HttpResponse('')


def fake_login(request):
    return HttpResponse('')


def django_exc(request):
    return get_object_or_404(Exception, pk=1)


def raise_exc(request):
    raise Exception(request.GET.get('message', 'view exception'))


def raise_ioerror(request):
    raise IOError(request.GET.get('message', 'view exception'))


def decorated_raise_exc(request):
    return raise_exc(request)


def template_exc(request):
    return render_to_response('error.html')


def ignored_exception(request):
    raise IgnoredException()


def logging_request_exc(request):
    logger = logging.getLogger(__name__)
    try:
        raise Exception(request.GET.get('message', 'view exception'))
    except Exception as e:
        logger.error(e, exc_info=True, extra={'request': request})
    return HttpResponse('')


def render_template_view(request):
    def something_expensive():
        with trace("something_expensive", "code"):
            return [User(username='Ron'), User(username='Beni')]

    return render(request, "list_users.html",
                            {'users': something_expensive})


def render_jinja2_template(request):
    return render(request, "jinja2_template.html")


def render_user_view(request):
    def something_expensive():
        with trace("something_expensive", "code"):
            for i in range(100):
                users = list(User.objects.all())
        return users

    return render(request, "list_users.html",
                  {'users': something_expensive})



"""
elasticapm.contrib.django.middleware
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import absolute_import

import logging
import threading

from django.conf import settings as django_settings

from elasticapm.contrib.django.models import client, get_client
from elasticapm.utils import (build_name_with_http_method_prefix,
                              disabled_due_to_debug, get_name_from_func, wrapt)

try:
    from importlib import import_module
except ImportError:
    from django.utils.importlib import import_module

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    # no-op class for Django < 1.10
    class MiddlewareMixin(object):
        pass


def _is_ignorable_404(uri):
    """
    Returns True if the given request *shouldn't* notify the site managers.
    """
    urls = getattr(django_settings, 'IGNORABLE_404_URLS', ())
    return any(pattern.search(uri) for pattern in urls)


class Catch404Middleware(MiddlewareMixin):
    def process_response(self, request, response):
        if (response.status_code != 404 or
                _is_ignorable_404(request.get_full_path())):
            return response
        if disabled_due_to_debug(
                    getattr(django_settings, 'ELASTICAPM', {}),
                    django_settings.DEBUG
                ):
            return response
        data = {
            'level': logging.INFO,
            'logger': 'http404',
        }
        result = client.capture(
            'Message',
            request=request,
            param_message={
                'message': 'Page Not Found: %s',
                'params': [request.build_absolute_uri()]
            }, data=data
        )
        request._elasticapm = {
            'app_name': data.get('app_name', client.app_name),
            'id': client.get_ident(result),
        }
        return response


def get_name_from_middleware(wrapped, instance):
    name = [type(instance).__name__, wrapped.__name__]
    if type(instance).__module__:
        name = [type(instance).__module__] + name
    return '.'.join(name)


def process_request_wrapper(wrapped, instance, args, kwargs):
    response = wrapped(*args, **kwargs)
    try:
        if response is not None:
            request = args[0]
            request._elasticapm_transaction_name = get_name_from_middleware(
                wrapped, instance
            )
    finally:
        return response


def process_response_wrapper(wrapped, instance, args, kwargs):
    response = wrapped(*args, **kwargs)
    try:
        request, original_response = args
        # if there's no view_func on the request, and this middleware created
        # a new response object, it's logged as the responsible transaction
        # name
        if (not hasattr(request, '_elasticapm_view_func')
                and response is not original_response):
            request._elasticapm_transaction_name = get_name_from_middleware(
                wrapped, instance
            )
    finally:
        return response


class TracingMiddleware(MiddlewareMixin):
    _elasticapm_instrumented = False
    _instrumenting_lock = threading.Lock()

    def __init__(self, *args, **kwargs):
        super(TracingMiddleware, self).__init__(*args, **kwargs)
        self.client = get_client()

        if not self._elasticapm_instrumented:
            with self._instrumenting_lock:
                if not self._elasticapm_instrumented:
                    if self.client.instrument_django_middleware:
                        self.instrument_middlewares()

                    TracingMiddleware._elasticapm_instrumented = True

    def instrument_middlewares(self):
        if getattr(django_settings, 'MIDDLEWARE_CLASSES', None):
            for middleware_path in django_settings.MIDDLEWARE_CLASSES:
                module_path, class_name = middleware_path.rsplit('.', 1)
                try:
                    module = import_module(module_path)
                    middleware_class = getattr(module, class_name)
                    if middleware_class == type(self):
                        # don't instrument ourselves
                        continue
                    if hasattr(middleware_class, 'process_request'):
                        wrapt.wrap_function_wrapper(
                            middleware_class,
                            'process_request',
                            process_request_wrapper,
                        )
                    if hasattr(middleware_class, 'process_response'):
                        wrapt.wrap_function_wrapper(
                            middleware_class,
                            'process_response',
                            process_response_wrapper,
                        )
                except ImportError:
                    client.logger.info(
                        "Can't instrument middleware %s", middleware_path
                    )

    def process_request(self, request):
        if not disabled_due_to_debug(
            getattr(django_settings, 'ELASTICAPM', {}),
            django_settings.DEBUG
        ):
            self.client.begin_transaction("request")

    def process_view(self, request, view_func, view_args, view_kwargs):
        request._elasticapm_view_func = view_func

    def process_response(self, request, response):
        try:
            if hasattr(response, 'status_code'):
                # check if _elasticapm_transaction_name is set
                if hasattr(request, '_elasticapm_transaction_name'):
                    transaction_name = request._elasticapm_transaction_name
                elif getattr(request, '_elasticapm_view_func', False):
                    transaction_name = get_name_from_func(
                        request._elasticapm_view_func
                    )
                else:
                    transaction_name = ''

                status_code = response.status_code
                transaction_name = build_name_with_http_method_prefix(
                    transaction_name,
                    request
                )
                transaction_data = self.client.get_data_from_request(request)
                self.client.set_transaction_extra_data(transaction_data,
                                                       'request')
                user_data = self.client.get_user_info(request)
                if user_data:
                    self.client.set_transaction_extra_data(
                        user_data, 'user')

                self.client.end_transaction(transaction_name, status_code)
        except Exception:
            self.client.error_logger.error(
                'Exception during timing of request',
                exc_info=True,
            )
        return response


class ErrorIdMiddleware(MiddlewareMixin):
    """
    Appends the X-ElasticAPM-ErrorId response header for referencing a message within
    the ElasticAPM datastore.
    """
    def process_response(self, request, response):
        if not getattr(request, '_elasticapm', None):
            return response
        response['X-ElasticAPM-ErrorId'] = request._elasticapm['id']
        return response


class LogMiddleware(MiddlewareMixin):
    # Create a thread local variable to store the session in for logging
    thread = threading.local()

    def process_request(self, request):
        self.thread.request = request



from __future__ import absolute_import

from django.conf import settings

from celery import Celery

from elasticapm.contrib.celery import register_signal
from elasticapm.contrib.django.models import client, logger, register_handlers

app = Celery('testapp')

app.config_from_object('django.conf:settings')
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)


# hook up ElasticAPM


try:
    register_signal(client)
except Exception as e:
    logger.exception('Failed installing celery hook: %s' % e)

if 'elasticapm.contrib.django' in settings.INSTALLED_APPS:
    register_handlers()



import functools
import logging
import os

from elasticapm.traces import get_transaction
from elasticapm.utils import wrapt
from elasticapm.utils.wrapt import resolve_path

logger = logging.getLogger(__name__)


# We have our own `BoundFunctionWrapper` and `FunctionWrapper` here because
# we want them to be able to now about `module` and `method` and supply it in
# the call to the wrapper.

class OriginalNamesBoundFunctionWrapper(wrapt.BoundFunctionWrapper):

    def __init__(self, *args, **kwargs):
        super(OriginalNamesBoundFunctionWrapper, self).__init__(*args, **kwargs)
        self._self_module = self._self_parent._self_module
        self._self_method = self._self_parent._self_method

    def __call__(self, *args, **kwargs):
        # If enabled has been specified, then evaluate it at this point
        # and if the wrapper is not to be executed, then simply return
        # the bound function rather than a bound wrapper for the bound
        # function. When evaluating enabled, if it is callable we call
        # it, otherwise we evaluate it as a boolean.

        if self._self_enabled is not None:
            if callable(self._self_enabled):
                if not self._self_enabled():
                    return self.__wrapped__(*args, **kwargs)
            elif not self._self_enabled:
                return self.__wrapped__(*args, **kwargs)

        # We need to do things different depending on whether we are
        # likely wrapping an instance method vs a static method or class
        # method.

        if self._self_binding == 'function':
            if self._self_instance is None:
                # This situation can occur where someone is calling the
                # instancemethod via the class type and passing the instance
                # as the first argument. We need to shift the args before
                # making the call to the wrapper and effectively bind the
                # instance to the wrapped function using a partial so the
                # wrapper doesn't see anything as being different.

                if not args:
                    raise TypeError(
                        'missing 1 required positional argument')

                instance, args = args[0], args[1:]
                wrapped = functools.partial(self.__wrapped__, instance)
                return self._self_wrapper(self._self_module,
                                          self._self_method,
                                          wrapped, instance, args, kwargs)

            return self._self_wrapper(self._self_module,
                                      self._self_method,
                                      self.__wrapped__, self._self_instance,
                                      args, kwargs)

        else:
            # As in this case we would be dealing with a classmethod or
            # staticmethod, then _self_instance will only tell us whether
            # when calling the classmethod or staticmethod they did it via an
            # instance of the class it is bound to and not the case where
            # done by the class type itself. We thus ignore _self_instance
            # and use the __self__ attribute of the bound function instead.
            # For a classmethod, this means instance will be the class type
            # and for a staticmethod it will be None. This is probably the
            # more useful thing we can pass through even though we loose
            # knowledge of whether they were called on the instance vs the
            # class type, as it reflects what they have available in the
            # decoratored function.

            instance = getattr(self.__wrapped__, '__self__', None)

            return self._self_wrapper(self._self_module,
                                      self._self_method,
                                      self.__wrapped__, instance, args,
                                      kwargs)

class OriginalNamesFunctionWrapper(wrapt.FunctionWrapper):

    __bound_function_wrapper__ = OriginalNamesBoundFunctionWrapper

    def __init__(self, wrapped, wrapper, module, method):
        super(OriginalNamesFunctionWrapper, self).__init__(wrapped, wrapper)
        self._self_module = module
        self._self_method = method

    def __call__(self, *args, **kwargs):
        # If enabled has been specified, then evaluate it at this point
        # and if the wrapper is not to be executed, then simply return
        # the bound function rather than a bound wrapper for the bound
        # function. When evaluating enabled, if it is callable we call
        # it, otherwise we evaluate it as a boolean.

        if self._self_enabled is not None:
            if callable(self._self_enabled):
                if not self._self_enabled():
                    return self.__wrapped__(*args, **kwargs)
            elif not self._self_enabled:
                return self.__wrapped__(*args, **kwargs)

        # This can occur where initial function wrapper was applied to
        # a function that was already bound to an instance. In that case
        # we want to extract the instance from the function and use it.

        if self._self_binding == 'function':
            if self._self_instance is None:
                instance = getattr(self.__wrapped__, '__self__', None)
                if instance is not None:
                    return self._self_wrapper(self._self_module,
                                              self._self_method,
                                              self.__wrapped__, instance,
                                              args, kwargs)

        # This is generally invoked when the wrapped function is being
        # called as a normal function and is not bound to a class as an
        # instance method. This is also invoked in the case where the
        # wrapped function was a method, but this wrapper was in turn
        # wrapped using the staticmethod decorator.

        return self._self_wrapper(self._self_module,
                                  self._self_method,
                                  self.__wrapped__, self._self_instance,
                                  args, kwargs)


class AbstractInstrumentedModule(object):
    name = None

    instrument_list = [
        # List of (module, method) pairs to instrument. E.g.:
        # ("requests.sessions", "Session.send"),
    ]

    def __init__(self):
        """

        :param client: elasticapm.base.Client
        """
        self.wrapped = None
        self.instrumented = False

        assert self.name is not None

    def get_wrapped_name(self, wrapped, instance, fallback_method=None):
        wrapped_name = []
        if(hasattr(instance, '__class__')
           and hasattr(instance.__class__, '__name__')):
            wrapped_name.append(instance.__class__.__name__)

        if hasattr(wrapped, '__name__'):
            wrapped_name.append(wrapped.__name__)
        elif fallback_method:
            attribute = fallback_method.split('.')
            if len(attribute) == 2:
                wrapped_name.append(attribute[1])

        return ".".join(wrapped_name)

    def get_instrument_list(self):
        return self.instrument_list

    def instrument(self):
        if self.instrumented:
            return

        skip_env_var = 'SKIP_INSTRUMENT_' + str(self.name.upper())
        if skip_env_var in os.environ:
            logger.debug("Skipping instrumentation of %s. %s is set.",
                         self.name, skip_env_var)
            return

        try:
            instrument_list = self.get_instrument_list()
            skipped_modules = set()

            for module, method in instrument_list:
                try:
                    # Skip modules we already failed to load
                    if module in skipped_modules:
                        continue
                    # We jump through hoop here to get the original
                    # `module`/`method` in the call to `call_if_sampling`
                    (parent, attribute, original) = resolve_path(module, method)
                    wrapper = OriginalNamesFunctionWrapper(
                        original,
                        self.call_if_sampling,
                        module,
                        method
                    )
                    wrapt.apply_patch(parent, attribute, wrapper)
                except ImportError:
                    # Could not import module
                    logger.debug("Skipping instrumentation of %s."
                                 " Module %s not found",
                                 self.name, module)

                    # Keep track of modules we couldn't load so we don't
                    # try to instrument anything in that module again
                    skipped_modules.add(module)
                except AttributeError as ex:
                    # Could not find thing in module
                    logger.debug("Skipping instrumentation of %s.%s: %s",
                                 module, method, ex)

        except ImportError as ex:
            logger.debug("Skipping instrumentation of %s. %s",
                         self.name, ex)
        self.instrumented = True

    def call_if_sampling(self, module, method, wrapped, instance, args, kwargs):
        if not get_transaction():
            return wrapped(*args, **kwargs)
        else:
            return self.call(module, method, wrapped, instance, args, kwargs)

    def call(self, module, method, wrapped, instance, args, kwargs):
        raise NotImplemented



"""
elasticapm.conf
~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import logging

__all__ = ('setup_logging', )


def setup_logging(handler, exclude=['elasticapm',
                                    'gunicorn',
                                    'south',
                                    'elasticapm.errors']):
    """
    Configures logging to pipe to Elastic APM.

    - ``exclude`` is a list of loggers that shouldn't go to ElasticAPM.

    For a typical Python install:

    >>> from elasticapm.handlers.logging import LoggingHandler
    >>> client = ElasticAPM(...)
    >>> setup_logging(LoggingHandler(client))

    Within Django:

    >>> from elasticapm.contrib.django.handlers import LoggingHandler
    >>> setup_logging(LoggingHandler())

    Returns a boolean based on if logging was configured or not.
    """
    logger = logging.getLogger()
    if handler.__class__ in map(type, logger.handlers):
        return False

    logger.addHandler(handler)

    # Add StreamHandler to sentry's default so you can catch missed exceptions
    for logger_name in exclude:
        logger = logging.getLogger(logger_name)
        logger.propagate = False
        logger.addHandler(logging.StreamHandler())

    return True



import os

from django.template.base import Node

from elasticapm.utils.stacks import get_frame_info

try:
    from django.template.base import Template
except ImportError:
    class Template(object):
        pass



def linebreak_iter(template_source):
    yield 0
    p = template_source.find('\n')
    while p >= 0:
        yield p + 1
        p = template_source.find('\n', p + 1)
    yield len(template_source) + 1


def get_data_from_template_source(source):
    origin, (start, end) = source
    template_source = origin.reload()

    lineno = None
    upto = 0
    source_lines = []
    for num, next in enumerate(linebreak_iter(template_source)):
        if start >= upto and end <= next:
            lineno = num
        source_lines.append(template_source[upto:next])
        upto = next

    if not source_lines or lineno is None:
        return {}

    pre_context = source_lines[max(lineno - 3, 0):lineno]
    post_context = source_lines[(lineno + 1):(lineno + 4)]
    context_line = source_lines[lineno]

    return {
        'template': {
            'filename': origin.loadname,
            'abs_path': origin.name,
            'pre_context': pre_context,
            'context_line': context_line,
            'lineno': lineno,
            'post_context': post_context,
        },
        'culprit': origin.loadname,
    }


def get_data_from_template_debug(template_debug):
    pre_context = []
    post_context = []
    context_line = None
    for lineno, line in template_debug['source_lines']:
        if lineno < template_debug['line']:
            pre_context.append(line)
        elif lineno > template_debug['line']:
            post_context.append(line)
        else:
            context_line = line
    return {
        'template': {
            'filename': os.path.basename(template_debug['name']),
            'abs_path': template_debug['name'],
            'pre_context': pre_context,
            'context_line': context_line,
            'lineno': template_debug['line'],
            'post_context': post_context,
        },
        'culprit': os.path.basename(template_debug['name']),
    }


def iterate_with_template_sources(frames, extended=True):
    template = None
    for frame, lineno in frames:
        f_code = getattr(frame, 'f_code', None)
        if f_code:
            function = frame.f_code.co_name
            if function == 'render':
                renderer = getattr(frame, 'f_locals', {}).get('self')
                if renderer and isinstance(renderer, Node):
                    if getattr(renderer, "token", None) is not None:
                        if hasattr(renderer, "source"):
                            # up to Django 1.8
                            yield {
                                'lineno': renderer.token.lineno,
                                'filename': renderer.source[0].name
                            }
                        else:
                            template = {'lineno': renderer.token.lineno}
                # Django 1.9 doesn't have the origin on the Node instance,
                # so we have to get it a bit further down the stack from the
                # Template instance
                elif renderer and isinstance(renderer, Template):
                    if template and getattr(renderer, 'origin', None):
                        template['filename'] = renderer.origin.name
                        yield template
                        template = None

        yield get_frame_info(frame, lineno, extended)



"""
elasticapm.utils.json_encoder
~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import datetime
import uuid

try:
    import json
except ImportError:
    import simplejson as json


class BetterJSONEncoder(json.JSONEncoder):
    ENCODERS = {
        set: list,
        frozenset: list,
        datetime.datetime: lambda obj: obj.strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
        uuid.UUID: lambda obj: obj.hex,
        bytes: lambda obj: obj.decode('utf-8', errors='replace')
    }

    def default(self, obj):
        if type(obj) in self.ENCODERS:
            return self.ENCODERS[type(obj)](obj)
        return super(BetterJSONEncoder, self).default(obj)


def better_decoder(data):
    return data


def dumps(value, **kwargs):
    return json.dumps(value, cls=BetterJSONEncoder, **kwargs)


def loads(value, **kwargs):
    return json.loads(value, object_hook=better_decoder)



import sys
from importlib import import_module

from elasticapm.utils import six


# From Django
# https://github.com/django/django/blob/master/django/utils/module_loading.py


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        msg = "%s doesn't look like a module path" % dotted_path
        six.reraise(ImportError, ImportError(msg), sys.exc_info()[2])

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        msg = 'Module "%s" does not define a "%s" attribute/class' % (
            module_path, class_name)
        six.reraise(ImportError, ImportError(msg), sys.exc_info()[2])



from __future__ import absolute_import

import logging

import mock

from elasticapm.conf import setup_logging
from tests.utils.compat import TestCase


class SetupLoggingTest(TestCase):
    def test_basic_not_configured(self):
        with mock.patch('logging.getLogger', spec=logging.getLogger) as getLogger:
            logger = getLogger()
            logger.handlers = []
            handler = mock.Mock()
            result = setup_logging(handler)
            self.assertTrue(result)

    def test_basic_already_configured(self):
        with mock.patch('logging.getLogger', spec=logging.getLogger) as getLogger:
            handler = mock.Mock()
            logger = getLogger()
            logger.handlers = [handler]
            result = setup_logging(handler)
            self.assertFalse(result)



#!/usr/bin/env python
"""
elasticapm
======

elastic-apm is a Python client for `Elastic APM <https://elastic.co/apm>`_. It provides
full out-of-the-box support for many of the popular frameworks, including
`Django <djangoproject.com>`_, `Flask <http://flask.pocoo.org/>`_, and `Pylons
<http://www.pylonsproject.org/>`_. elastic-apm also includes drop-in support for any
`WSGI <http://wsgi.readthedocs.org/>`_-compatible web application.
"""

# Hack to prevent stupid "TypeError: 'NoneType' object is not callable" error
# in multiprocessing/util.py _exit_function when running `python
# setup.py test` (see
# http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html)
for m in ('multiprocessing', 'billiard'):
    try:
        __import__(m)
    except ImportError:
        pass

import sys
import os
from codecs import open

from setuptools import setup, find_packages, Extension
from elasticapm.version import VERSION
from setuptools.command.test import test as TestCommand

from distutils.command.build_ext import build_ext
from distutils.errors import (CCompilerError, DistutilsExecError,
                              DistutilsPlatformError)

if sys.platform == 'win32':
    build_ext_errors = (CCompilerError, DistutilsExecError,
                        DistutilsPlatformError, IOError)
else:
    build_ext_errors = (CCompilerError, DistutilsExecError,
                        DistutilsPlatformError)


class BuildExtFailed(Exception):
    pass


class optional_build_ext(build_ext):
    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            raise BuildExtFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except build_ext_errors:
            raise BuildExtFailed()

tests_require = [
    'py>=1.4.26',
    'pytest>=2.6.4',
    'pytest-django==2.8.0',
    'pytest-capturelog>=0.7',
    'blinker>=1.1',
    'celery',
    'django-celery',
    'Flask>=0.8',
    'logbook',
    'mock',
    'pep8',
    'webob',
    'pytz',
    'redis',
    'requests',
    'jinja2',
    'pytest-benchmark',
    'urllib3-mock',
    'Twisted',

    # isort
    'apipkg',
    'execnet',
    'isort',
    'pytest-cache',
    'pytest-isort',
]

if sys.version_info[0] == 2:
    tests_require += [
        'unittest2',
        'gevent',
        'zerorpc>=0.4.0,<0.5',
        'python-memcached'
    ]
else:
    tests_require += ['python3-memcached']


if sys.version_info[:2] == (2, 6):
    tests_require += ['Django>=1.2,<1.7']
else:
    tests_require += ['Django>=1.2']

try:
    import __pypy__
except ImportError:
    if sys.version_info[0] == 2 and 'SKIP_ZERORPC' not in os.environ:
        tests_require += ['zerorpc>=0.4.0,<0.5']
    tests_require += ['psycopg2']

if sys.version_info >= (3, 5):
    tests_require += [
        'aiohttp',
        'pytest-asyncio',
        'pytest-mock',
    ]

install_requires = ['urllib3', 'certifi']

try:
    # For Python >= 2.6
    import json
except ImportError:
    install_requires.append("simplejson>=2.3.0,<2.5.0")


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup_kwargs = dict(
    name='elastic-apm',
    version=VERSION,
    author='Elastic, Inc',
    author_email='support@opbeat.com',
    url='https://github.com/elastic/apm-agent-python',
    description='The official Python module for Elastic APM',
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.rst'), encoding='utf-8').read(),
    packages=find_packages(exclude=("tests",)),
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={'tests': tests_require,
                    'flask': ['blinker'],
                    'asyncio': ['aiohttp']},
    cmdclass={'test': PyTest},
    test_suite='tests',
    include_package_data=True,
    entry_points={
        'paste.filter_app_factory': [
            'elasticapm = elasticapm.contrib.paste:filter_factory',
        ],
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Topic :: Software Development',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)


def run_setup(with_extensions):
    setup_kwargs_tmp = dict(setup_kwargs)

    if with_extensions:
        setup_kwargs_tmp['ext_modules'] = [Extension(
            'elasticapm.utils.wrapt._wrappers', ['elasticapm/utils/wrapt/_wrappers.c']
        )]
        setup_kwargs_tmp['cmdclass']['build_ext'] = optional_build_ext

    setup(**setup_kwargs_tmp)

# Figure out if we should build the wrapt C extensions

with_extensions = os.environ.get('ELASTICAPM_WRAPT_EXTENSIONS', None)

if with_extensions:
    if with_extensions.lower() == 'true':
        with_extensions = True
    elif with_extensions.lower() == 'false':
        with_extensions = False
    else:
        with_extensions = None

if hasattr(sys, 'pypy_version_info'):
    with_extensions = False

if with_extensions is None:
    with_extensions = True

try:
    run_setup(with_extensions=with_extensions)
except BuildExtFailed:
    run_setup(with_extensions=False)



# -*- coding: utf-8 -*-


class InvalidScheme(ValueError):
    """
    Raised when a transport is constructed using a URI which is not
    handled by the transport
    """


class DuplicateScheme(Exception):
    """
    Raised when registering a handler for a particular scheme which
    is already registered
    """
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import datetime
import uuid

from elasticapm.utils import json_encoder as json
from elasticapm.utils import six
from tests.utils.compat import TestCase


class JSONTest(TestCase):
    def test_uuid(self):
        res = uuid.uuid4()
        self.assertEquals(json.dumps(res), '"%s"' % res.hex)

    def test_datetime(self):
        res = datetime.datetime(day=1, month=1, year=2011, hour=1, minute=1, second=1)
        self.assertEquals(json.dumps(res), '"2011-01-01T01:01:01.000000Z"')

    def test_set(self):
        res = set(['foo', 'bar'])
        self.assertIn(json.dumps(res), ('["foo", "bar"]', '["bar", "foo"]'))

    def test_frozenset(self):
        res = frozenset(['foo', 'bar'])
        self.assertIn(json.dumps(res), ('["foo", "bar"]', '["bar", "foo"]'))

    def test_bytes(self):
        if six.PY2:
            res = bytes('foobar')
        else:
            res = bytes('foobar', encoding='ascii')
        self.assertEqual(json.dumps(res), '"foobar"')



import pytest  # isort:skip
pytest.importorskip("flask")  # isort:skip

import mock
from flask import Flask, render_template, signals

from elasticapm.contrib.flask import ElasticAPM
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


def create_app():
    app = Flask(__name__)

    @app.route('/an-error/', methods=['GET', 'POST'])
    def an_error():
        raise ValueError('hello world')

    @app.route('/users/', methods=['GET', 'POST'])
    def users():
        return render_template('users.html',
                               users=['Ron', 'Rasmus'])

    return app


class FlaskTest(TestCase):
    def setUp(self):
        self.app = create_app()
        self.client = self.app.test_client()

        self.elasticapm_client = get_tempstoreclient()
        self.elasticapm = ElasticAPM(self.app, client=self.elasticapm_client)

    def tearDown(self):
        signals.request_started.disconnect(self.elasticapm.request_started)
        signals.request_finished.disconnect(self.elasticapm.request_finished)

    def test_error_handler(self):
        response = self.client.get('/an-error/')
        self.assertEquals(response.status_code, 500)
        self.assertEquals(len(self.elasticapm_client.events), 1)

        event = self.elasticapm_client.events.pop(0)['errors'][0]

        self.assertTrue('exception' in event)
        exc = event['exception']
        self.assertEquals(exc['type'], 'ValueError')
        self.assertEquals(exc['message'], 'ValueError: hello world')
        self.assertEquals(event['culprit'], 'tests.contrib.flask.flask_tests.an_error')

    def test_get(self):
        response = self.client.get('/an-error/?foo=bar')
        self.assertEquals(response.status_code, 500)
        self.assertEquals(len(self.elasticapm_client.events), 1)

        event = self.elasticapm_client.events.pop(0)['errors'][0]

        self.assertTrue('request' in event['context'])
        request = event['context']['request']
        self.assertEquals(request['url']['raw'], 'http://localhost/an-error/?foo=bar')
        self.assertEquals(request['url']['search'], 'foo=bar')
        self.assertEquals(request['method'], 'GET')
        self.assertEquals(request['body'], None)
        self.assertTrue('headers' in request)
        headers = request['headers']
        self.assertTrue('content-length' in headers, headers.keys())
        self.assertEquals(headers['content-length'], '0')
        self.assertTrue('content-type' in headers, headers.keys())
        self.assertEquals(headers['content-type'], '')
        self.assertTrue('host' in headers, headers.keys())
        self.assertEquals(headers['host'], 'localhost')
        env = request['env']
        self.assertTrue('SERVER_NAME' in env, env.keys())
        self.assertEquals(env['SERVER_NAME'], 'localhost')
        self.assertTrue('SERVER_PORT' in env, env.keys())
        self.assertEquals(env['SERVER_PORT'], '80')

    def test_get_debug(self):
        self.app.config['DEBUG'] = True
        self.app.config['TESTING'] = False
        self.assertRaises(ValueError, self.app.test_client().get, '/an-error/?foo=bar')
        self.assertEquals(len(self.elasticapm_client.events), 0)

    def test_get_debug_elasticapm(self):
        self.app.config['DEBUG'] = True
        self.app.config['TESTING'] = True
        self.app.config['ELASTICAPM'] = {'DEBUG': True}
        self.assertRaises(ValueError, self.app.test_client().get, '/an-error/?foo=bar')
        self.assertEquals(len(self.elasticapm_client.events), 1)

    def test_post(self):
        response = self.client.post('/an-error/?biz=baz', data={'foo': 'bar'})
        self.assertEquals(response.status_code, 500)
        self.assertEquals(len(self.elasticapm_client.events), 1)

        event = self.elasticapm_client.events.pop(0)['errors'][0]

        self.assertTrue('request' in event['context'])
        request = event['context']['request']
        self.assertEquals(request['url']['raw'], 'http://localhost/an-error/?biz=baz')
        self.assertEquals(request['url']['search'], 'biz=baz')
        self.assertEquals(request['method'], 'POST')
        self.assertEquals(request['body'], 'foo=bar')
        self.assertTrue('headers' in request)
        headers = request['headers']
        self.assertTrue('content-length' in headers, headers.keys())
        self.assertEquals(headers['content-length'], '7')
        self.assertTrue('content-type' in headers, headers.keys())
        self.assertEquals(headers['content-type'], 'application/x-www-form-urlencoded')
        self.assertTrue('host' in headers, headers.keys())
        self.assertEquals(headers['host'], 'localhost')
        env = request['env']
        self.assertTrue('SERVER_NAME' in env, env.keys())
        self.assertEquals(env['SERVER_NAME'], 'localhost')
        self.assertTrue('SERVER_PORT' in env, env.keys())
        self.assertEquals(env['SERVER_PORT'], '80')

    def test_instrumentation(self):
        with mock.patch("elasticapm.traces.TransactionsStore.should_collect") as should_collect:
            should_collect.return_value = False
            resp = self.client.post('/users/')

        assert resp.status_code == 200, resp.response

        transactions = self.elasticapm_client.instrumentation_store.get_all()

        # If the test falls right at the change from one minute to another
        # this will have two items.
        assert len(transactions) == 1
        traces = transactions[0]['traces']
        assert len(traces) == 2, [t['name'] for t in traces]

        expected_signatures = ['transaction', 'users.html']

        assert set([t['name'] for t in traces]) == set(expected_signatures)

        assert traces[1]['name'] == 'transaction'
        assert traces[1]['type'] == 'transaction'

        assert traces[0]['name'] == 'users.html'
        assert traces[0]['type'] == 'template.jinja2'

    def test_instrumentation_404(self):
        with mock.patch("elasticapm.traces.TransactionsStore.should_collect") as should_collect:
            should_collect.return_value = False
            resp = self.client.post('/no-such-page/')

        assert resp.status_code == 404, resp.response

        transactions = self.elasticapm_client.instrumentation_store.get_all()

        expected_signatures = ['transaction']

        assert len(transactions) == 1
        traces = transactions[0]['traces']
        print(transactions[0])
        assert transactions[0]['result'] == '404'
        assert len(traces) == 1, [t["signature"] for t in traces]

        assert set([t['name'] for t in traces]) == set(expected_signatures)

        assert traces[0]['name'] == 'transaction'
        assert traces[0]['type'] == 'transaction'

    def test_framework_version(self):
        elasticapm = ElasticAPM(app=self.app)
        app_info = elasticapm.client.get_app_info()
        assert 'flask' == app_info['framework']['name']



from elasticapm.instrumentation.packages.dbapi2 import Literal, scan, tokenize


def test_scan_simple():
    sql = "Hello 'Peter Pan' at Disney World"
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", Literal("'", "Peter Pan"), "at", "Disney", "World"]
    assert actual == expected


def test_scan_with_escape_single_quote():
    sql = "Hello 'Peter\\' Pan' at Disney World"
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", Literal("'", "Peter' Pan"), "at", "Disney", "World"]
    assert actual == expected


def test_scan_with_escape_slash():
    sql = "Hello 'Peter Pan\\\\' at Disney World"
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", Literal("'", "Peter Pan\\"), "at", "Disney", "World"]
    assert actual == expected


def test_scan_double_quotes():
    sql = """Hello 'Peter'' Pan''' at Disney World"""
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", Literal("'", "Peter' Pan'"), "at", "Disney", "World"]
    assert actual == expected


def test_scan_double_quotes_at_end():
    sql = """Hello Peter Pan at Disney 'World'"""
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", "Peter", "Pan", "at", "Disney", Literal("'", "World")]
    assert actual == expected



"""
elasticapm.utils.encoding
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import uuid

from elasticapm.utils import six


def is_protected_type(obj):
    """Determine if the object instance is of a protected type.

    Objects of protected types are preserved as-is when passed to
    force_text(strings_only=True).
    """
    from decimal import Decimal
    import datetime
    return isinstance(obj, six.integer_types + (type(None), float, Decimal,
        datetime.datetime, datetime.date, datetime.time))


def force_text(s, encoding='utf-8', strings_only=False, errors='strict'):
    """
    Similar to smart_text, except that lazy instances are resolved to
    strings, rather than kept as lazy objects.

    If strings_only is True, don't convert (some) non-string-like objects.
    """
    # Handle the common case first, saves 30-40% when s is an instance of
    # six.text_type. This function gets called often in that setting.
    #
    # Adapted from Django
    if isinstance(s, six.text_type):
        return s
    if strings_only and is_protected_type(s):
        return s
    try:
        if not isinstance(s, six.string_types):
            if hasattr(s, '__unicode__'):
                s = s.__unicode__()
            else:
                if six.PY3:
                    if isinstance(s, bytes):
                        s = six.text_type(s, encoding, errors)
                    else:
                        s = six.text_type(s)
                else:
                    s = six.text_type(bytes(s), encoding, errors)
        else:
            # Note: We use .decode() here, instead of six.text_type(s, encoding,
            # errors), so that if s is a SafeBytes, it ends up being a
            # SafeText at the end.
            s = s.decode(encoding, errors)
    except UnicodeDecodeError as e:
        if not isinstance(s, Exception):
            raise UnicodeDecodeError(*e.args)
        else:
            # If we get to here, the caller has passed in an Exception
            # subclass populated with non-ASCII bytestring data without a
            # working unicode method. Try to handle this without raising a
            # further exception by individually forcing the exception args
            # to unicode.
            s = ' '.join([force_text(arg, encoding, strings_only,
                    errors) for arg in s])
    return s


def _has_elasticapm_metadata(value):
    try:
        return callable(value.__getattribute__("__elasticapm__"))
    except:
        return False


def transform(value, stack=None, context=None):
    # TODO: make this extendable
    if context is None:
        context = {}
    if stack is None:
        stack = []

    objid = id(value)
    if objid in context:
        return '<...>'

    context[objid] = 1
    transform_rec = lambda o: transform(o, stack + [value], context)

    if any(value is s for s in stack):
        ret = 'cycle'
    elif isinstance(value, (tuple, list, set, frozenset)):
        try:
            ret = type(value)(transform_rec(o) for o in value)
        except Exception:
            # We may be dealing with a namedtuple
            class value_type(list):
                __name__ = type(value).__name__
            ret = value_type(transform_rec(o) for o in value)
    elif isinstance(value, uuid.UUID):
        ret = repr(value)
    elif isinstance(value, dict):
        ret = dict((to_unicode(k), transform_rec(v)) for k, v in six.iteritems(value))
    elif isinstance(value, six.text_type):
        ret = to_unicode(value)
    elif isinstance(value, six.binary_type):
        ret = to_string(value)
    elif not isinstance(value, six.class_types) and \
            _has_elasticapm_metadata(value):
        ret = transform_rec(value.__elasticapm__())
    elif isinstance(value, bool):
        ret = bool(value)
    elif isinstance(value, float):
        ret = float(value)
    elif isinstance(value, int):
        ret = int(value)
    elif six.PY2 and isinstance(value, long):
        ret = long(value)
    elif value is not None:
        try:
            ret = transform(repr(value))
        except:
            # It's common case that a model's __unicode__ definition may try to query the database
            # which if it was not cleaned up correctly, would hit a transaction aborted exception
            ret = u'<BadRepr: %s>' % type(value)
    else:
        ret = None
    del context[objid]
    return ret


def to_unicode(value):
    try:
        value = six.text_type(force_text(value))
    except (UnicodeEncodeError, UnicodeDecodeError):
        value = '(Error decoding value)'
    except Exception:  # in some cases we get a different exception
        try:
            value = six.binary_type(repr(type(value)))
        except Exception:
            value = '(Error decoding value)'
    return value


def to_string(value):
    try:
        return six.binary_type(value.decode('utf-8').encode('utf-8'))
    except:
        return to_unicode(value).encode('utf-8')


def shorten(var, list_length=50, string_length=200):
    var = transform(var)
    if isinstance(var, six.string_types) and len(var) > string_length:
        var = var[:string_length - 3] + '...'
    elif isinstance(var, (list, tuple, set, frozenset)) and len(var) > list_length:
        # TODO: we should write a real API for storing some metadata with vars when
        # we get around to doing ref storage
        # TODO: when we finish the above, we should also implement this for dicts
        var = list(var)[:list_length] + ['...', '(%d more elements)' % (len(var) - list_length,)]
    return var



"""
elasticapm.contrib.django.models
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Acts as an implicit hook for Django installs.

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import absolute_import

import logging
import os
import sys
import warnings

from django.conf import settings as django_settings

import elasticapm.instrumentation.control
from elasticapm.utils import disabled_due_to_debug, six

logger = logging.getLogger('elasticapm.errors.client')


def get_installed_apps():
    """
    Generate a list of modules in settings.INSTALLED_APPS.
    """
    out = set()
    for app in django_settings.INSTALLED_APPS:
        out.add(app)
    return out

_client = (None, None)


class ProxyClient(object):
    """
    A proxy which represents the currenty client at all times.
    """
    # introspection support:
    __members__ = property(lambda x: x.__dir__())

    # Need to pretend to be the wrapped class, for the sake of objects that care
    # about this (especially in equality tests)
    __class__ = property(lambda x: get_client().__class__)

    __dict__ = property(lambda o: get_client().__dict__)

    __repr__ = lambda: repr(get_client())
    __getattr__ = lambda x, o: getattr(get_client(), o)
    __setattr__ = lambda x, o, v: setattr(get_client(), o, v)
    __delattr__ = lambda x, o: delattr(get_client(), o)

    __lt__ = lambda x, o: get_client() < o
    __le__ = lambda x, o: get_client() <= o
    __eq__ = lambda x, o: get_client() == o
    __ne__ = lambda x, o: get_client() != o
    __gt__ = lambda x, o: get_client() > o
    __ge__ = lambda x, o: get_client() >= o
    if six.PY2:
        __cmp__ = lambda x, o: cmp(get_client(), o)
    __hash__ = lambda x: hash(get_client())
    # attributes are currently not callable
    # __call__ = lambda x, *a, **kw: get_client()(*a, **kw)
    __nonzero__ = lambda x: bool(get_client())
    __len__ = lambda x: len(get_client())
    __getitem__ = lambda x, i: get_client()[i]
    __iter__ = lambda x: iter(get_client())
    __contains__ = lambda x, i: i in get_client()
    __getslice__ = lambda x, i, j: get_client()[i:j]
    __add__ = lambda x, o: get_client() + o
    __sub__ = lambda x, o: get_client() - o
    __mul__ = lambda x, o: get_client() * o
    __floordiv__ = lambda x, o: get_client() // o
    __mod__ = lambda x, o: get_client() % o
    __divmod__ = lambda x, o: get_client().__divmod__(o)
    __pow__ = lambda x, o: get_client() ** o
    __lshift__ = lambda x, o: get_client() << o
    __rshift__ = lambda x, o: get_client() >> o
    __and__ = lambda x, o: get_client() & o
    __xor__ = lambda x, o: get_client() ^ o
    __or__ = lambda x, o: get_client() | o
    __div__ = lambda x, o: get_client().__div__(o)
    __truediv__ = lambda x, o: get_client().__truediv__(o)
    __neg__ = lambda x: -(get_client())
    __pos__ = lambda x: +(get_client())
    __abs__ = lambda x: abs(get_client())
    __invert__ = lambda x: ~(get_client())
    __complex__ = lambda x: complex(get_client())
    __int__ = lambda x: int(get_client())
    if not six.PY2:
        __long__ = lambda x: long(get_client())
    __float__ = lambda x: float(get_client())
    __str__ = lambda x: str(get_client())
    __unicode__ = lambda x: six.text_type(get_client())
    __oct__ = lambda x: oct(get_client())
    __hex__ = lambda x: hex(get_client())
    __index__ = lambda x: get_client().__index__()
    __coerce__ = lambda x, o: x.__coerce__(x, o)
    __enter__ = lambda x: x.__enter__()
    __exit__ = lambda x, *a, **kw: x.__exit__(*a, **kw)

client = ProxyClient()

default_client_class = 'elasticapm.contrib.django.DjangoClient'


def get_client_class(client_path=default_client_class):
    module, class_name = client_path.rsplit('.', 1)
    return getattr(__import__(module, {}, {}, class_name), class_name)


def get_client_config():
    config = getattr(django_settings, 'ELASTICAPM', {})
    if 'ASYNC' in config:
        warnings.warn(
            'Usage of "ASYNC" configuration is deprecated. Use "ASYNC_MODE"',
            category=DeprecationWarning,
            stacklevel=2,
        )
        config['ASYNC_MODE'] = config['ASYNC']
    return dict(
        servers=config.get('SERVERS', None),
        include_paths=set(
            config.get('INCLUDE_PATHS', [])) | get_installed_apps(),
        exclude_paths=config.get('EXCLUDE_PATHS', None),
        filter_exception_types=config.get('FILTER_EXCEPTION_TYPES', None),
        timeout=config.get('TIMEOUT', None),
        hostname=config.get('HOSTNAME', None),
        auto_log_stacks=config.get('AUTO_LOG_STACKS', None),
        string_max_length=config.get('MAX_LENGTH_STRING', None),
        list_max_length=config.get('MAX_LENGTH_LIST', None),
        app_name=config.get('APP_NAME', None),
        secret_token=config.get('SECRET_TOKEN', None),
        transport_class=config.get('TRANSPORT_CLASS', None),
        processors=config.get('PROCESSORS', None),
        traces_send_freq_secs=config.get('TRACES_SEND_FREQ_SEC', None),
        async_mode=config.get('ASYNC_MODE', None),
        instrument_django_middleware=config.get('INSTRUMENT_DJANGO_MIDDLEWARE'),
        transactions_ignore_patterns=config.get('TRANSACTIONS_IGNORE_PATTERNS', None),
    )


def get_client(client=None):
    """
    Get an ElasticAPM client.

    :param client:
    :return:
    :rtype: elasticapm.base.Client
    """
    global _client

    tmp_client = client is not None
    if not tmp_client:
        config = getattr(django_settings, 'ELASTICAPM', {})
        client = config.get('CLIENT', default_client_class)

    if _client[0] != client:
        client_class = get_client_class(client)
        instance = client_class(**get_client_config())
        if not tmp_client:
            _client = (client, instance)
        return instance
    return _client[1]


def exception_handler(request=None, **kwargs):
    def actually_do_stuff(request=None, **kwargs):
        exc_info = sys.exc_info()
        try:
            if (
                disabled_due_to_debug(
                    getattr(django_settings, 'ELASTICAPM', {}),
                    django_settings.DEBUG
                )
                or getattr(exc_info[1], 'skip_elasticapm', False)
            ):
                return

            get_client().capture('Exception', exc_info=exc_info,
                                 request=request)
        except Exception as exc:
            try:
                logger.exception(u'Unable to process log entry: %s' % (exc,))
            except Exception as exc:
                warnings.warn(u'Unable to process log entry: %s' % (exc,))
        finally:
            try:
                del exc_info
            except Exception as e:
                logger.exception(e)

    return actually_do_stuff(request, **kwargs)


def register_handlers():
    from django.core.signals import got_request_exception

    # Connect to Django's internal signal handler
    got_request_exception.connect(exception_handler)

    # If Celery is installed, register a signal handler
    if 'djcelery' in django_settings.INSTALLED_APPS:
        from elasticapm.contrib.celery import register_signal

        try:
            register_signal(client)
        except Exception as e:
            logger.exception('Failed installing django-celery hook: %s' % e)

    # Instrument to get traces
    skip_env_var = 'SKIP_INSTRUMENT'
    if skip_env_var in os.environ:
        logger.debug("Skipping instrumentation. %s is set.", skip_env_var)
    else:
        elasticapm.instrumentation.control.instrument()


if 'elasticapm.contrib.django' in django_settings.INSTALLED_APPS:
    register_handlers()



"""
elasticapm.conf.defaults
~~~~~~~~~~~~~~~~~~~

Represents the default values for all ElasticAPM settings.

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import os
import os.path
import socket

ROOT = os.path.normpath(os.path.join(os.path.dirname(__file__), os.pardir))

# Allow local testing of ElasticAPM even if DEBUG is enabled
DEBUG = False

# This should be the schema+host of the APM Server server
SERVERS = ['http://localhost:8080']

# Error API path
ERROR_API_PATH = '/v1/errors'

# Transactions API path
TRANSACTIONS_API_PATH = '/v1/transactions'

TIMEOUT = 20

# TODO: this is specific to Django
CLIENT = 'elasticapm.contrib.django.DjangoClient'

HOSTNAME = socket.gethostname()

# Credentials to authenticate with the APM Server server
ACCESS_TOKEN = None

# Extending this allow you to ignore module prefixes when we attempt to
# discover which function an error comes from (typically a view)
EXCLUDE_PATHS = []

# By default ElasticAPM only looks at modules in INSTALLED_APPS for drilling down
# where an exception is located
INCLUDE_PATHS = []

# The maximum number of elements to store for a list-like structure.
MAX_LENGTH_LIST = 50

# The maximum length to store of a string-like structure.
MAX_LENGTH_STRING = 400

MAX_LENGTH_VALUES = {
    'message':        200,
    'server_name':    200,
    'culprit':        100,
    'logger':         60
}

# Automatically log frame stacks from all ``logging`` messages.
AUTO_LOG_STACKS = False

# Client-side data processors to apply
PROCESSORS = (
    'elasticapm.processors.sanitize_stacktrace_locals',
    'elasticapm.processors.sanitize_http_request_cookies',
    'elasticapm.processors.sanitize_http_request_headers',
    'elasticapm.processors.sanitize_http_wsgi_env',
    'elasticapm.processors.sanitize_http_request_querystring',
    'elasticapm.processors.sanitize_http_request_body',
)

# How often we send data to the metrics backend
TRACES_SEND_FREQ_SECS = 60

# Should data be sent to the APM Server asynchronously in a separate thread
ASYNC_MODE = True

# Should elasticapm wrap middleware for better metrics detection
INSTRUMENT_DJANGO_MIDDLEWARE = True

SYNC_TRANSPORT_CLASS = 'elasticapm.transport.http_urllib3.Urllib3Transport'

ASYNC_TRANSPORT_CLASS = 'elasticapm.transport.http_urllib3.AsyncUrllib3Transport'

TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'



import datetime
import os

import pymongo
import pytest

import elasticapm
import elasticapm.instrumentation.control
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


@pytest.mark.integrationtest
class InstrumentPyMongoTest(TestCase):
    def setUp(self):
        self.client = get_tempstoreclient()
        elasticapm.instrumentation.control.instrument()
        connection_params = {'host': os.environ.get('MONGODB_HOST', 'localhost'),
                             'port': int(os.environ.get('MONGODB_PORT', 27017))}
        if pymongo.version_tuple < (3, 0):
            connection_params['safe'] = True
        self.mongo = pymongo.MongoClient(**connection_params)
        self.db = self.mongo.elasticapm_test

    def tearDown(self):
        self.mongo.drop_database('elasticapm_test')

    @pytest.mark.skipif(pymongo.version_tuple < (3, 0), reason='New in 3.0')
    def test_collection_bulk_write(self):
        self.client.begin_transaction('transaction.test')
        requests = [pymongo.InsertOne({'x': 1}),
                    pymongo.DeleteOne({'x': 1}),
                    pymongo.ReplaceOne({'w': 1}, {'z': 1}, upsert=True)]
        result = self.db.blogposts.bulk_write(requests)
        self.assertEqual(result.inserted_count, 1)
        self.assertEqual(result.deleted_count, 1)
        self.assertEqual(result.upserted_count, 1)
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.bulk_write')

    def test_collection_count(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        self.db.blogposts.insert(blogpost)
        self.client.instrumentation_store.get_all()
        self.client.begin_transaction('transaction.test')
        count = self.db.blogposts.count()
        self.assertEqual(count, 1)
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'], 'elasticapm_test.blogposts.count')

    @pytest.mark.skipif(pymongo.version_tuple < (3, 0), reason='New in 3.0')
    def test_collection_delete_one(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        self.db.blogposts.insert_one(blogpost)
        self.client.begin_transaction('transaction.test')
        r = self.db.blogposts.delete_one({'author': 'Tom'})
        self.assertEqual(r.deleted_count, 1)
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.delete_one')

    @pytest.mark.skipif(pymongo.version_tuple < (3, 0), reason='New in 3.0')
    def test_collection_delete_many(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        self.db.blogposts.insert_one(blogpost)
        self.client.begin_transaction('transaction.test')
        r = self.db.blogposts.delete_many({'author': 'Tom'})
        self.assertEqual(r.deleted_count, 1)
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.delete_many')

    def test_collection_insert(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        self.client.begin_transaction('transaction.test')
        r = self.db.blogposts.insert(blogpost)
        self.assertIsNotNone(r)
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.insert')

    @pytest.mark.skipif(pymongo.version_tuple < (3, 0), reason='New in 3.0')
    def test_collection_insert_one(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        self.client.begin_transaction('transaction.test')
        r = self.db.blogposts.insert_one(blogpost)
        self.assertIsNotNone(r.inserted_id)
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.insert_one')

    @pytest.mark.skipif(pymongo.version_tuple < (3, 0), reason='New in 3.0')
    def test_collection_insert_many(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        self.client.begin_transaction('transaction.test')
        r = self.db.blogposts.insert_many([blogpost])
        self.assertEqual(len(r.inserted_ids), 1)
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()

        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.insert_many')

    def test_collection_find(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        blogposts = []
        for i in range(1000):
            blogposts.append({'author': 'Tom', 'comments': i})
        self.db.blogposts.insert(blogposts)
        r = self.db.blogposts.insert(blogpost)
        self.client.instrumentation_store.get_all()
        self.client.begin_transaction('transaction.test')
        r = list(self.db.blogposts.find({'comments': {'$gt': 995}}))

        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.cursor.refresh')

    @pytest.mark.skipif(pymongo.version_tuple < (3, 0), reason='New in 3.0')
    def test_collection_find_one(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        r = self.db.blogposts.insert_one(blogpost)
        self.client.begin_transaction('transaction.test')
        r = self.db.blogposts.find_one({'author': 'Tom'})
        self.assertEqual(r['author'], 'Tom')
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.find_one')

    def test_collection_remove(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        r = self.db.blogposts.insert(blogpost)
        self.client.begin_transaction('transaction.test')
        r = self.db.blogposts.remove({'author': 'Tom'})
        self.assertEqual(r['n'], 1)
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.remove')

    def test_collection_update(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        r = self.db.blogposts.insert(blogpost)
        self.client.begin_transaction('transaction.test')
        r = self.db.blogposts.update({'author': 'Tom'},
                                     {'$set': {'author': 'Jerry'}})
        self.assertEqual(r['n'], 1)
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.update')

    @pytest.mark.skipif(pymongo.version_tuple < (3, 0), reason='New in 3.0')
    def test_collection_update_one(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        r = self.db.blogposts.insert(blogpost)
        self.client.begin_transaction('transaction.test')
        r = self.db.blogposts.update_one({'author': 'Tom'},
                                     {'$set': {'author': 'Jerry'}})
        self.assertEqual(r.modified_count, 1)
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.update_one')

    @pytest.mark.skipif(pymongo.version_tuple < (3, 0), reason='New in 3.0')
    def test_collection_update_many(self):
        blogpost = {'author': 'Tom', 'text': 'Foo',
                    'date': datetime.datetime.utcnow()}
        r = self.db.blogposts.insert(blogpost)
        self.client.begin_transaction('transaction.test')
        r = self.db.blogposts.update_many({'author': 'Tom'},
                                     {'$set': {'author': 'Jerry'}})
        self.assertEqual(r.modified_count, 1)
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.blogposts.update_many')

    @pytest.mark.skipif(pymongo.version_tuple < (2, 7), reason='New in 2.7')
    def test_bulk_execute(self):
        self.client.begin_transaction('transaction.test')
        bulk = self.db.test_bulk.initialize_ordered_bulk_op()
        bulk.insert({'x': 'y'})
        bulk.insert({'z': 'x'})
        bulk.find({'x': 'y'}).replace_one({'x': 'z'})
        bulk.execute()
        self.client.end_transaction('transaction.test')
        transactions = self.client.instrumentation_store.get_all()
        trace = _get_pymongo_trace(transactions[0]['traces'])
        self.assertEqual(trace['type'], 'db.mongodb.query')
        self.assertEqual(trace['name'],
                         'elasticapm_test.test_bulk.bulk.execute')


def _get_pymongo_trace(traces):
    for trace in traces:
        if trace['type'].startswith('db.mongodb'):
            return trace
